import math, random, copy
import numpy as np 
import os, os.path as pathmap
from rmidi import mutils
import itertools
from rmidi.constant import converter, meta_event_format, ch_event_format, sys_event_format, Constants

numpy = np

WRAP_DATA = 0x7f
WRAP_BITS = 7

def ch_event_id(id, no) :
    return ((id & 0xf) << 4) + (no & 0xf)

def metype(typ):
    return typ & 0xff

def to_var_length(k):
    if k > 127:
        leng = length(k) // WRAP_BITS + 1
        var = bytearray(leng)
        var[-1] = k & WRAP_DATA
        k >>= WRAP_BITS
        for i in range(leng - 2, -1, -1):
            var[i] = (k & WRAP_DATA) + WRAP_DATA + 1
            k >>= WRAP_BITS
        return var
    else:
        return bytearray((k,))     

def to_fix_length(k, leng, bits):
    fix = bytearray(leng)
    wrapper = (1 << bits) - 1
    if k > 255 and k > -1:
        fix[-1] = k & wrapper
        k >>= bits
        for i in range(leng - 2, -1, -1):
            fix[i] = (k & wrapper) + wrapper + 1
            k >>= bits
    elif k > -1 :
        fix[-1] = k & 0xff
    return fix

def up(n_b, base = 2):
    return base ** int(math.log2(n_b) / math.log2(base) + 1)
    
def split(n, t, block_size = None): #Splits 'n' integer in t integer of bits bit(n)/t
    n_b = length(n)
    sbit = up(n_b) // t if not block_size else block_size
    WRAPPER = 2 ** sbit - 1
    k = n
    li = []
    while k:
        li.append(k & WRAPPER)
        k >>= sbit
    le = t - len(li)
    for _ in range(le):
        li.append(0)
    li.reverse()
    return li


def merge(a, b, *nums, block_size = 16):#Merge the numbers in order
    WRAPPER = (1 << block_size) - 1
    num = ((a & WRAPPER) << block_size) + (b & WRAPPER)
    for v in nums:
        num = (num << block_size) + (v & WRAPPER)
    return num


def length(k):
    """Finds length of k in bits
    
    Arguments:
        k {int} -- Integer number
    """
    return int(math.log2(k) + 1)
def find_location(text, listt):
    try:
        return next((i, j) 
            for i, t in enumerate(listt)
            for j, v in enumerate(t)
            if v == text)
    except StopIteration:
        return None

class Constant:
    Mthd = (0x4d, 0x54, 0x68, 0x64)
    Mtrk = (0x4d, 0x54, 0x72, 0x6b)
    META_EVENT_END = (0x01, 0xff, 0x2f, 0x0)
    MAX = 1048576
    DTYPE = {'uint4': (0, 16), 'uint7': (0, 127), 'uint8' : (0, 256)}
    CHANNEL_EVENT = 'CHANNEL'
    META_EVENT = 'META'
    SYS_EVENT = 'SYS'
    #format = (id, sub_event_name, param_len, params..., param_type)
    ch_event_format = (
                        (0x8, "note_off", 2, "note_number", "velocity", "int"),
                        (0x9, "note_on", 2, "note_number", "velocity", "int"),
                        (0xA, "after_touch", 2, "note_number", "amount", "int"),
                        (0xB, "controller", 2, "controller_type", "value", "int"),                   
                        (0xC, "program_change", 1, "program_number", "int"),                   
                        (0xD, "channel_after_touch", 1, "amount", "int"),                   
                        (0xE, "pitch_bend", 2, "vlsb", "vmsb", "int")                  
                    )
    #format = (id, "sub_event_name", "param_len", params..., "param_type", "common_range, -1", //if -1, highest_allowed_val_of_param..., -1, "mask")
    #If -1 in place of common range the following param len args are highesh values of parameters
    #If all within mask positive else masked negative
    meta_event_format = (
                            (0x00, "sequence_number", 2, "nmsb", "nlsb", "int", 256),
                            (0x01, "text_event", -1, "text", "str"),
                            (0x02, "copyright_notice", -1, "text", "str"),
                            (0x03, "track_name", -1, "text", "str"),
                            (0x04, "instrument_name", -1, "text", "str"),
                            (0x05, "lyrics", -1, "text", "str"),
                            (0x06, "marker", -1, "text", "str"),
                            (0x07, "cue_point", -1, "text", "str"),
                            (0x20, "midi_ch_prefix", 1, "channel", "int", 16),
							(0x21, "midi_port", 1, "port_no", "int", 16),
                            (0x2F, "end_of_track", 0, None, "None"),
                            (0x51, "set_tempo", 3, "musec_per_quat_note", "int", 8355712, 3),
                            (0x54, "smpte_offset", 5, "hr", "min", "sec", "fr", "subfr", "int", -1, 24, 60, 60, 30, 100),
                            (0x58, "time_sig", 4, "numer", "denom", "metro", "32nds", "int", 255),
                            (0x59, "key_sig", 2, "key", "scale", "int" ,-1, 14, 2, -1, 7),
                            (0x7F, "sequence_specifier", -1, "data", "any")
                        )
    sys_event_format = (
                            (0xF0, "normal_sys_event", -1, "data", "int", 128),
                            (0xF7, "authorization_sys_event", -1, "data", "int", 256),
                            
                        )

    #controller format = (controller_value/id , "controller_type", "len_of_simmilar_extension")
    controller = (
                    (0x00, "blank_select", 1),
                    (0x01, "modulation", 1),
                    (0x02, "breath_controller", 1),
                    (0x04, "foot_controller", 1),
                    (0x05, "portamento_time", 1),
                    (0x06, "data_entry", 1),#May be msb
                    (0x07, "main_volume", 1),
                    (0x08, "balance", 1),
                    (0x0A, "pan", 1),
                    (0x0B, "expression_controller", 1),
                    (0x0C, "effect_control", 2),
                    # (0x0D, "effect_control_2", 0),
                    (0x10, "general_purpose_controller", 4, 0x50, 4),
                    (0x20, "LSB_for_controller", 32),
                    (0x40, "damper_pedal", 1),
                    (0x41, "portamento", 1),
                    (0x42, "sostenuto", 1),
                    (0x43, "soft_pedal", 1),
                    (0x44, "legato_footswitch", 1),
                    (0x45, "hold2", 1),
                    (0x46, "damper_pedal", 1),
                    (0x47, "sound_controller", 10),
                    (0x40, "damper_pedal", 1),
                    (0x50, "post_general_purpose_controller", 4),
                    (0x54, "portamento_control", 1),
                    (0x5B, "effects_depth", 5),
                    (0x60, "data_increment", 1),
                    (0x61, "data_decrement", 1),
                    (0x62, "nonreg_param_num_lsb", 1),
                    (0x63, "nonreg_param_num_msb", 1),
                    (0x64, "reg_param_num_lsb", 1),
                    (0x65, "reg_param_num_msb", 1),
                    (0x79, "mode_messages", 7),
                )


class MIDI:
    def __init__(self,  format_type = 0, track_count = 0, time_div = 0x1e0, empty = False, filename = None, ):
        # return super().__init__(*args, **kwargs)
        self.pipe = {}
        self.filename = filename
        self.format_type = format_type
        self.track_count = track_count
        self.time_div = time_div
        self.tracks = [MIDI.Track(self, empty= empty) for _ in range(track_count)]
        self.byte_list = bytearray()     

    def track(self, track_no):#Indexing from Zero
        if self.track_count < track_no : raise IndexError("Track no out of range")
        return self.tracks[track_no]
    def set_time_div(self, val, inbpm = True): 
        """val acts as actual bpm value if inbpm is True, 
         or as no ticks per seconds count if inbpm is False
         It actually sets ticks resolution
        Arguments:
            self {MIDI} -- [description]
            val {int} 
        Keyword Arguments:
            inbpm {bool} -- [description] (default: {True})
        """
        val = (val * .5e6 ) // 7200 if inbpm else val 
        self.pipe['ntime_div']  = val
        self.time_div = val

    def to_byte_array(self):
        byte_list = self.__midiheaderbytes()

        for trk in self.tracks:
            byte_list.extend(trk.to_byte_array())
        return byte_list
    
    def create_file(self, file_name, loc = pathmap.abspath(os.curdir)):
        with open(loc + "/" + file_name + ".mid", "wb+") as f:
            df = self.to_byte_array()

            f.write(df)

    @classmethod
    def parse_midi(cls, filename):
        with open(filename, 'rb') as f:
            content = f.read()
            # print(cont)
            return MIDI.parser(content)
        return None

    @classmethod
    def parser(cls, content):
        try : 
            vmthd = content[:4]
            vlength = mutils.toint(content[4:8])
            vformat_type = mutils.toint(content[8:10])
            vtrack_count = mutils.toint(content[10:12])
            vtime_div = mutils.toint(content[12:14])

            mid = cls(vformat_type, vtrack_count, vtime_div, empty = True)
            utape = 14
            for t in mid.tracks:
                vmtrk = content[utape: utape + 4]
                lentr = mutils.toint(content[utape + 4: utape + 8])
                utape += (lentr + 8)
                trkip = content[utape - lentr : utape]
                leny = len(trkip)
                if leny == 0: break
                trk_elist = []
                to, fr = 0, 0 #Start state
                tape = -1 #Start at far left

                vbuf = bytearray()
                del_time = 0
                kparams = {}
                # print('trk :\n', mutils.hexstr(trkip, leng= 16, group=2, numlen= 2))
                # hex_string = "".join("%02x" % b for b in content)
                # print("E--> ", hex_string)
                # print(trkip)
                while True:
                    if tape > leny: 
                        raise ValueError("File has sys event, not supported by us : rmidi")

                    tape += 1
                    # print('tape ---> ', tape)
                    # if tape < leny : print('state =: ', to , ', tape ---> :', hex(tape), ',  i/p : ', hex(trkip[tape]))
                    if to == 0: #Init State, Deltatime Detection
                        vbuf.append(trkip[tape])
                        if trkip[tape] < 0x80:
                            del_time = mutils.vartoint(vbuf)
                            vbuf = bytearray()
                            to = 1

                    elif to == 1: #Event Detection
                        # print('In event Detection L : ')
                        if 0x80 <= trkip[tape] < 0xF0: to = 2 #Channel Event
                        elif trkip[tape] == 0xFF: to = 7 #Meta Event 
                        elif trkip[tape] == 0xF0 or trkip[tape] == 0xF7 : to = 4
                        elif 'fromevent' in kparams:
                            if kparams['fromevent'] == 'Channel':
                                tape -= 1
                                to = 5
                            else : raise ValueError('kparams wrongly set , attr - fromevent is ' + kparams['fromevent']) 
                        else : raise Exception('Error in Event detection phase : no event matched ... event id : ' + hex(trkip[tape]))
                    
                    elif to == 2: #Channel Event
                        tape -= 1
                        ev = mutils.numin(trkip[tape], 4, 4)
                        ind = find_location(ev, Constant.ch_event_format)
                        if ind:
                            evt_info = Constant.ch_event_format[ind[0]]
                            kparams['event_id'] = trkip[tape]
                            kparams['len'] = evt_info[2]
                            to = 5
                            pass
                        else : raise Exception("Can't parse the file\nChannel Event not in parsers list\nChannel Event ID: " + hex(trkip[tape]))
                        fr = 2
                        
                    elif to == 3: #Meta Event
                            if 'mlength' in kparams:
                                # print("le = ", hex(kparams.get('le')), ", ip len : ", hex(kparams.get('mlength')))
                                if kparams.get('le') == 0:
                                    trk_elist.append(MIDI.Track.Event.MetaEvent(del_time, stype, params= ()))
                                    break
                                elif kparams.get('le') == kparams.get('mlength') or kparams.get('le') == -1:
                                    tape += kparams.get('mlength')
                                    params = trkip[tape - le: tape]
                                    trk_elist.append(MIDI.Track.Event.MetaEvent(del_time, stype, params= params))
                                    tape -= 1
                                else: raise ValueError("Event length from file cant match with list avail.")
                            else: raise Exception('kparams errors... mlength not in kparams')

                            kparams['fromevent'] = 'Meta'
                            to = 0
                            
                
                    elif to == 4: #Sys Event
                        to = 6
                        kparams['fromevent'] = 'sys'
                        pass
                    elif to == 5: #Sub Channel Event
                        tape += kparams.get('len') 
                        params = trkip[tape - kparams.get('len') : tape]
                        tape -= 1
                        # print('params : ',mutils.hexstr(params))
                        trk_elist.append(MIDI.Track.Event.ChannelEvent(del_time, kparams.get('event_id'), params= params))
                        kparams['fromevent'] = 'Channel'
                        to = 0

                    elif to == 6: #Var length calculator
                        vbuf.append(trkip[tape])
                        if trkip[tape] < 0x80:
                            kparams['mlength'] = mutils.vartoint(vbuf)
                            vbuf = bytearray()
                            if 'fromevent' in kparams:
                                if kparams['fromevent'] == 'sys':
                                    to = 8
                                    continue
                            to = 3
                        pass
                    elif to == 7: #Meta sub event
                        ind = find_location(trkip[tape], Constant.meta_event_format)
                        if ind:
                            evt_info = Constant.meta_event_format[ind[0]]
                            stype = evt_info[0]
                            le = evt_info[2]
                            kparams['le'] = le
                            kparams['stype'] = stype
                            to = 6
                        else : raise Exception(str(e) + "\nCan't parse the file\nMeta Event Subtype not in parsers list\nMeta Event Subtype ID: " + hex(trkip[tape]))
                    elif to == 8:
                        if 'mlength' in kparams:
                            tape += kparams.get('mlength')
                        to == 0


                t.__reinit__(id, len(trk_elist), trk_elist, mid)
        except Exception as e:
            raise Exception('Something not alright : rimidi' + str(e))
        # print(mid)
        return mid
    def decompress(self):
        if self.filename is not None:
            return MIDI.parse_midi(self.filename)

    def compress(self, filename = 'default'):
        bytelist = self.__midiheaderbytes()
        # print("H : ", mutils.hexstr(bytelist))
        ch_active = False
        for t in self.tracks:
            bytelist.extend(Constant.Mtrk)
            start = len(bytelist)
            eventbuff = bytearray()
            for e in t.trk_event:
                eventbuff = e.to_byte_array()
                if e.etype == Constant.CHANNEL_EVENT:
                    if ch_active == e.event_id: eventbuff.remove(e.event_id)
                    ch_active = e.event_id
                    pass
                else : ch_active = False
                bytelist.extend(eventbuff)
            le = mutils.to_fix_length(len(bytelist) - start, 4, 8)
            for i in range(start, start + 4):
                bytelist.insert(i, le[i - start - 4])
        with open(filename + '.mid','wb+') as f:
            f.write(bytelist)
        return bytelist


    """
    Fucntion for all tracks, 
    """

    def transpose_to(self, scale, s_type = 0, new = False):
        self_copy = copy.deepcopy(self)
        for t in self_copy.tracks:
            t.transpose_to(scale, s_type)
        return self_copy

    def change_scale(self, scale, s_type = 0, new = False):
        self_copy = copy.deepcopy(self)
        for t in self_copy.tracks:
            t.change_scale(scale)
        return self_copy

    def note_frequencies(self):
        freq_count = numpy.zeros((self.track_count, 256))
        for t in range(self.track_count):
            freq_count[t] = self.tracks[t].note_frequencies()
        return freq_count

    def __midiheaderbytes(self):
        byte_list = bytearray()
        byte_list.extend(Constant.Mthd)
        byte_list.extend(mutils.to_fix_length(6, 4, 8))
        byte_list.extend(mutils.to_fix_length(self.format_type, 2, 8))
        byte_list.extend(mutils.to_fix_length(self.track_count, 2, 8))
        byte_list.extend(mutils.to_fix_length(self.time_div, 2, 8))
        return byte_list
    
    def __refresh__(self):
        for t in self.tracks:
            pass

    def __repr__(self):
        st = ""
        for t in self.tracks:
            st = st + t.__repr__() + "\n" 
        return st

    class Track:
        def __init__(self, midi, empty = False):
            self.id = 0
            self.length = 0
            self.trk_event = []
            self.midi_ = midi
            # self.track_m = self
            if empty : return
            self.trk_event.append(MIDI.Track.Event.MetaEvent(0, 0x58,(0x4, 0x2, 0x18, 0x8))) #Meta Event : Time Signature
            self.trk_event.append(MIDI.Track.Event.MetaEvent(0, 0x59, (0x0, 0x0))) #Meta Event : Key Signature
            self.trk_event.append(MIDI.Track.Event.MetaEvent(0, 0x51, (0x7, 0xa1, 0x20)))
            self.trk_event.append(MIDI.Track.Event.ChannelEvent(0, 0xb, 0x0, (0x79, 0))) #Channel Event : Controller  Blank Select
            self.trk_event.append(MIDI.Track.Event.ChannelEvent(0, 0xc, 0, (0x0,))) #Channel Event : Program Change Event
            self.trk_event.append(MIDI.Track.Event.ChannelEvent(0, 0xb, 0, (0x7, 0x64))) #Channel Event : Controller  Main Volume
            self.trk_event.append(MIDI.Track.Event.ChannelEvent(0, 0xb, 0, (0x0a, 0x40))) #Channel Event : Controller  Pan
            self.trk_event.append(MIDI.Track.Event.ChannelEvent(0, 0xb, 0, (0x5b, 0x00)))#/Channel Event : Controller  External Effects Depth
            self.trk_event.append(MIDI.Track.Event.ChannelEvent(0, 0xb, 0, (0x5d, 0x00))) #Channel Event : Controller  formerly Chorus Depth
            self.trk_event.append(MIDI.Track.Event.MetaEvent(0, 0x04, params="Piano")) # Intrument name
            self.trk_event.append(MIDI.Track.Event.MetaEvent(0, 0x21,(0x00,))) # Meta Event : Midi Port/ Should be prior to anyone
            
            self.trk_event.append(MIDI.Track.Event.MetaEvent(0, 0x2f, None)) #Meta Event : End of Track
        
        def __reinit__(self, id, length, trk_event, midi_):
            self.id = id
            self.length = length
            trk_event_copy = copy.deepcopy(trk_event)
            # print("Trakc : ... ", trk_event)
            # print("Track Length : ", len(trk_event))
            self.trk_event = [v for v in trk_event_copy]
            # print("Trak : ... ", self.trk_event)
            self.midi_ = midi_

        def _add_event(self, evt):
            end = self.trk_event.pop()
            self.trk_event.append(evt)
            self.trk_event.append(end)
            return None

           

        def set_tempo(self, time, tempo):
            tmp = ((500000 * 120) // tempo)
            self.add_event(0, 'set_tempo', tempo = tmp)

        def tempo(self):
            return mutils.toint(self.get_event('set_tempo', depth = 1)[0].data, 8)

        def transpose_to(self, scale, s_type = 0, offset = -256, new = False):
            self_copy = copy.deepcopy(self) if new else self
            if offset != -256: pass

            elif type(scale) == int and -7 < scale < 7 : 
                scale_id, s_type = self.get_scale(string= False)
                scale = Constants.SCALE[scale_id + s_type & 12]
            elif scale in Constants.SCALE:
                 scale_id, s_type = self.get_scale()
                 offset = (Constants.SCALE_OFFSET[scale_id] - Constants.SCALE_OFFSET[scale])  
            else : raise AttributeError("Scale not defined. ")
            
            for e in self_copy.trk_event:
                if e.is_note_on_off_event():
                    e.data[0] -= offset
            return self_copy

        def transpose(self, tval):
            arr = self.notes()
            return [(arr[0], arr[1] + tval, arr[2]) for v in arr]

        def change_scale(self, scale, octave = 1, new = False):
            """Change the scale of midi track,  
            Assumes the c-major has no flats instead no sharps;  key sig event will    | 0xFF      |  0x59      |         0x00            |            0x01            |
                                                                                        meta marker  key-sig       no of sharps or flats      flat or sharp indicator  
            
            Arguments:
                scale {string} -- scale name in sharp notation format, like c-major, a#-minor

            
            Keyword Arguments:
                new {bool} -- if change scale on new duplicated track (default: {False})
                ocatave{int} -- relative integer indicating to tranfer or scale change in back or ahead octave {default: {1}}
            """
            self_copy = copy.deepcopy(self) if new else self
            if scale in Constants.SCALE_OFFSET:
                scale_id, s_type = self.get_scale()

                key_sig_event = self.get_event('key_sig', new = False, depth = 1)[0]
                key_sig = [(Constants.SCALE_KEY_SIG[scale][0] & 7) ,  mutils.invert(Constants.SCALE_KEY_SIG[scale][1] & 7, 8, True)]
                key_sig = [key_sig[1], 1] if key_sig[0] == 0 else [key_sig[0], 0] # convert  the key sig tuple to (data, type) tuple as in midi format
                key_sig_event.set_date_params(key_sig)

                transpose_int = (Constants.SCALE_OFFSET[scale_id] - Constants.SCALE_OFFSET[scale]) - (octave - 1) * 12
                self.transpose_to(0, offset= transpose_int)
                return key_sig_event
            return None

        def get_scale(self, string = True, sharps = False):
            """Finds the scale, or pattern in which track is. Can only find major, minor scale for now 
            
            Keyword Arguments:
                string {bool} -- True if wants string - representation of scale(SCALE NAME) (default: {True})
                ifmajor {int} -- [description] (default: {0})
            
            Raises:
                NotImplementedError: [description]
            
            Returns:
                [type] -- [description]
            """
            # event_info = mutils.find_in_nested_dict(Constants.meta_event_format_dict, 'key_sig')
            event_info = Constants.meta_event_format_dict[0xFF][0x59] # 0x59 <---> key_sig
            sharps = 1 if sharps else 0
            for e in self.trk_event:
                # e =  Event()
                # if e.is_meta_event() : print(e)
                if e.meta_event_type == 0x59: #  0x59 <---> key_sig, key signature event
                    key, scale = mutils.magnitude(e.data[0], 8), e.data[1]
                    if string: return Constants.SCALE_KEY_SIG_REV[key | ((sharps & 1) | 8)][sharps], scale    
                    return key | ((scale & 1) | 8), scale
                    # raise NotImplementedError("Implement the get_scale to use it.") 
            if string: return 'c-major', 0
            return 0, 0
        def get_event(self, event, depth = 10000, new = True):
            res = []
            ind1 = find_location(event, Constant.ch_event_format)
            dep = 0
            if ind1:
                id_ = Constant.ch_event_format[ind1[0]][0]
                for e in self.trk_event:
                    if e.event_id == id_: 
                        # res.append(copy.deepcopy(e))
                        if new : res.append(copy.deepcopy(e))
                        else : res.append(e)
                    dep += 1
                    if dep > depth: break
                return res            
            ind2 = find_location(event, Constant.meta_event_format)
            dep = 0
            if ind2:
                id_ = Constant.meta_event_format[ind2[0]][0]
                event_info = meta_event_format.X[0xFF][id_]
                # event_info = event_info['0xFF'][mutils.hex2(id_)]
                for e in self.trk_event:
                    if e.meta_event_type == id_: 
                        if new : res.append(copy.deepcopy(e))
                        else : res.append(e)
                        dep += 1
                        if dep > depth: break
                if id_ in {0x4} and len(res) == 0: 
                    ne = MIDI.Track.Event.MetaEvent(0, id_, event_info['default'])
                    return [ne]
                return res

            # ind = ind1 if not ind1  else ind2 if not ind2  else None
            # print("ind : From get_event --  ", ind)
            # if not ind:
            #     print("id_ : From get_event -- ", id_)
                
        def add_event(self, time, event, **kparams):#Can't add key sig event for now
            if 'delta' in kparams:
                pass
            else:
                if type(event) == str:
                    #Adding channel event
                    ind = find_location(event, Constant.ch_event_format)
                    # if event in itertools.chain(*Constant.ch_event_format):
                    if ind:
                        evt_info = Constant.ch_event_format[ind[0]]
                        ch_no = kparams.get('channel_no') if 'channel_no' in kparams else 0
                        event_id = (evt_info[0] << 4) | (ch_no & 7)
                        if 'params' in kparams:
                            params = kparams.get('params')
                            if len(params) != evt_info[2]: raise ValueError("params length not matched to actual parameter")
                            pass
                        else:
                            params = [0] * evt_info[2]
                            for i in range(3, 3 + evt_info[2]):
                                if  evt_info[i] in kparams:
                                    params[i - 3] = kparams.get(evt_info[i])
                                    pass
                                else: raise ValueError("Attribute " + evt_info[i] + " not in parameters")
                        self._add_event(MIDI.Track.Event(delta_time= self.delta_time(time), etype= Constant.CHANNEL_EVENT, event_id= event_id, data=bytearray(params)))
                        return True
                    #Adding meta event    
                    ind = find_location(event, Constant.meta_event_format)
                    if ind :
                        evt_info = Constant.meta_event_format[ind[0]]
                        event_id = 0xff
                        meta_event_type = evt_info[0]
                        if event == 'set_tempo':
                            a, b, c = split(kparams.get('tempo'), 3, 8)
                            self._add_event(MIDI.Track.Event(delta_time=self.delta_time(time), etype= Constant.META_EVENT, event_id= event_id, meta_event_type= meta_event_type, data= bytearray((a, b, c))))
                            return
                        if evt_info[2] == -1:
                            if 'text' in kparams:
                                if type(kparams.get('text')) == str:
                                    params = kparams.get('text').encode('utf-8')
                                else :raise ValueError("Event : " + event + " text attribute is nor string")
                            elif 'data' in kparams:
                                if type(kparams.get('data')) == str:
                                    params = kparams.get('data').encode('utf-8')
                                else :raise ValueError("Event : " + event + " text attribute is nor string")
                        elif evt_info[2] == 0: self.trk_event.append(MIDI.Track.Event.MetaEvent(0, 0x2f, None)) #Meta Event : End of Track
                        else :
                            if evt_info[3+ evt_info[2]] == -1:
                                if 'key_sig' in kparams:
                                    return
                                maxr = [evt_info[i] for i in range(4 + evt_info[2], 4 + 2 * evt_info[2])]
                            else : maxr = [evt_info[3+ evt_info[2]]] * evt_info[2]
                            # print(maxr)
                            for i in range(3, 3 + evt_info[2]):
                                params = [0] * evt_info[2]
                                if evt_info[i] in kparams:
                                    params[i - 3] = kparams.get(evt_info[i]) % maxr[i]
                                    pass
                                else : raise ValueError("Attribute " + evt_info[i] + " not in parameters")

                        self._add_event(MIDI.Track.Event(delta_time= self.delta_time(time), etype= Constant.META_EVENT, event_id= event_id, meta_event_type= meta_event_type, data=bytearray(params)))

                        pass
                    if event in itertools.chain(*Constant.ch_event_format):
                        pass
                else:
                    raise ValueError("Inappropiate Arhument Value, event not of type string")


            return None
        def push_note(self, note_length, note_val, channel_no = 0, intensity = 0x50):
            """Pushes the note on self track, over midi object
            
            Arguments:
                note_length {int} -- tradition note length, whole(1), half(2), quater(4) and so on
                note_val {int} -- Corrseponding note value, related to midi encoding
            
            Keyword Arguments:
                channel_no {int} -- midi channel , necessity for channel (default: {0})
                intensity {hexadecimal} -- loudness, volume, or how hard the note is hit (default: {0x50})
            """
            end = self.trk_event.pop()
            self.trk_event.append(MIDI.Track.Event.ChannelEvent(self.delta_time(note_length), 0x9, channel_no, (note_val, intensity)))
            self.trk_event.append(end)


        def close_note(self, note_length, note_val, channel_no, intensity = 0):
            """Pushes the note on self track, over midi object
            
            Arguments:
                note_length {int} -- tradition note length, whole(1), half(2), quater(4) and so on
                note_val {int} -- Corrseponding note value, related to midi encoding
            
            Keyword Arguments:
                channel_no {int} -- midi channel , necessity for channel (default: {0})
                intensity {hexadecimal} -- loudness, volume, or how hard the note is hit (default: {0x50})
            """
            end = self.trk_event.pop()
            self.trk_event.append(MIDI.Track.Event.ChannelEvent(self.delta_time(note_length), 0x8, channel_no, (note_val, 0x30)))
            self.trk_event.append(end)
        
        def delta_time(self, note_length):#Computes the delta time from tradition note duration, full, half, quater 
            """ Computes the delta time from tradition note duration, full, half, quater 
            Assumes every midi file of 4 / 4 key sig for now 
            
            Arguments:
                note_length {int} -- tradition note duration, full(1), half(2), quater(4)
            Returns:
                [int] -- delta_time
            """
            if not note_length: return 0
            return (self.midi_.time_div * 4) // note_length
        
        def leng(self):#Returns length in bytes
            szn = 0
            for e in self.trk_event:
                szn += e.leng()
            return szn
        
        def to_byte_array(self):
            byte_list = bytearray()
            byte_list.extend(Constant.Mtrk)
            # byte_list.append([0] * 4)
            for e in self.trk_event:
                byte_list.extend(e.to_byte_array())
            le = mutils.to_fix_length(self.leng(), 4, 8)
            for i in range(4, 8):
                byte_list.insert(i, le[i - 4])
            return byte_list

        def note_frequencies(self):
            freq_count = numpy.zeros(256) #index represent midi note
            for e in self.trk_event:
                # e = Event()
                if e.is_note_off_event():
                    freq_count[e.data[0]] += 1 #incerementin note count of data[0]
            return freq_count

        def min_note(self):
            fre = numpy.trim_zeros(self.note_frequencies())
            return fre[0]

            

        def notes(self, abs = True, eventtype = None):
            dictn = {}
            res = []
            res_bpm = [(0, 500000),]
            timekeeper = 0 #in microsecods
            for t in self.trk_event:
                timekeeper += (t.delta_time / self.midi_.time_div * res_bpm[-1][1])
                if 0x80 <= t.event_id < 0xa0:   
                    eventid = t.event_id >> 4
                    noteval = t.data[0]
                    velocity = t.data[1] 
                    to = 0
                    #DFA Inside
                    if to == 0:#Delta time == 0
                        to = 1 
                    if to == 1: #check if closing the node
                        if (eventid  & 0xf == 8) or velocity == 0: to = 3
                        else : to = 2
                    if to == 2: # push the current time
                        dictn[noteval] = timekeeper
                        to = -1
                    if to == 3: # assign the duration
                        duration = timekeeper - dictn[noteval]
                        to = 4
                    if to == 4: # store in res
                        res.append(((timekeeper - duration)/1e6, noteval, duration/1e6))
                        
                elif t.event_id == 0xff and t.meta_event_type == 0x51:
                    microsecods = t.get_data()
                    if (timekeeper, microsecods) not in res_bpm: res_bpm += [(timekeeper, microsecods)]
            return {'note_series' : res, 'bpm_change' : res_bpm}  

        def __refresh__(self):
            for t in self.trk_event:
                pass

        def __repr__(self):
            st = ""
            for e in self.trk_event:
                st = st +  e.__repr__() + "\n"
            return st


        class Event :
            
            def __init__(self, delta_time = 0, etype = None, event_id = None, meta_event_type = None, length = 0, data = bytearray()):
                self.delta_time = delta_time
                self.etype = etype
                self.event_id = event_id
                self.meta_event_type = meta_event_type
                self.length = length
                self.data = data
                self.bytes = bytearray()
                self.abstime = 0
                self.elength = 0

            def set_date_params(self, params):
                self.data = bytearray(params)

            def scale_delta_time(self, factor):
                self.delta_time = self.delta_time * factor

            def get_data(self, asnum = True, msb = True):
                return mutils.toint(self.data, 8)
            
            def is_channel_event(self):
                return self.etype == Constant.CHANNEL_EVENT

            def is_meta_event(self):
                return self.etype == Constant.META_EVENT

            def is_sys_event(self):
                return self.etype == Constant.SYS_EVENT
            def is_note_on_off_event(self):
                return 0x7f < self.event_id < 0xa0

            def is_note_off_event(self):
                return 0x8f < self.event_id < 0x90 or (False if len(self.data) < 2 else self.data[1] == 0) #checked in note channel id == 0x8 or velocity == 0
                
            @classmethod
            def ChannelEvent(cls, delta_time, event_id, channel_no = None, params = ()):
                evti = mutils.ch_event_id(event_id, channel_no) if channel_no is not None else event_id 
                # print("Ch event  evti : ", evti, ", None : ", channel_no is None)
                evt = cls(delta_time, Constant.CHANNEL_EVENT, evti)
                evt.data = bytearray(params)
                return evt

            
            @classmethod
            def MetaEvent(cls, delta_time, meta_event_type, params = ()):
                evt = cls(delta_time, Constant.META_EVENT, 0xff, meta_event_type & 0xff)
                if params is None : return evt
                elif type(params) == str :                    
                    evt.data = params.encode('utf-8')
                else : evt.data = bytearray(params)
                return evt

            @classmethod
            def SysEvent(cls, delta_time, event_id, params = ()):
                evt = cls(delta_time, Constant.SYS_EVENT, event_id & 0xff)
                evt.data = bytearray(params)
                return evt
            
            def to_byte_array(self):
                byte_list = bytearray()
                byte_list.extend(mutils.to_var_length(self.delta_time))
                byte_list.append(self.event_id)
                if self.etype == Constant.CHANNEL_EVENT:
                    # print("ch : ", self.event_id)
                    # byte_list.pop()
                    # evti = mutils.ch_event_id(self.event_id, self.channel_no)
                    # byte_list.append(evti)
                    pass
                elif self.etype == Constant.META_EVENT:
                    byte_list.append(self.meta_event_type)
                    byte_list.extend(mutils.to_var_length(len(self.data)))
                    pass
                elif self.etype == Constant.SYS_EVENT:
                    byte_list.extend(mutils.to_var_length(len(self.data)))
                    pass
                # print(self.data)
                byte_list.extend(self.data)
                self.bytes.extend(byte_list)
                self.length = len(self.bytes)
                # hex_string = "".join("%02x" % b for b in byte_list)
                # print("E--> ", hex_string)
                return byte_list
            def leng(self, data = True):
                if data : return len(self.data)
                return len(self.to_byte_array())
            
            def __repr__(self):
                # print("E--> ", hex_string)
                # mhxmet = hex(self.meta_event_type) if self.meta_event_type else '0'
                # return "Delta Time : " + hex(self.delta_time) + ", Etype : " + self.etype + ", Event ID : " + hex(self.event_id) + ", META : " + mhxmet + ", Length : " + hex(self.leng()) + ", Data --> " + mutils.hexstr(self.data, group= 2)
                mstr = '______________________________________________________________________________________________________________________________ . . . \n'
                mstr += '| Absolute Time   |  Duration       |  Delta Time |  ETYPE     |   Event ID | META  | LENGTH     | DATA \n' #% (e.abstime, e.elength, hex(e.delta_time), e.etype, hex(e.event_id), mhxmet, hex(e.leng()),  mutils.hexstr(e.data, group= 2))
                mstr += '|______________________________________________________________________________________________________________________________ . . . \n'
                mhxmet = hex(self.meta_event_type) if self.meta_event_type else '0'
                # estr = " Abs Time : " + str(e.abstime) + "Duration : " + str(e.elength) + " |||>>>> Delta Time : " + hex(e.delta_time) + ", Etype : " + e.etype + ", Event ID : " + hex(e.event_id) + ", META : " + mhxmet + ", Length : " + hex(e.leng()) + ", Data --> " + mutils.hexstr(e.data, group= 2)
                estr = '| %-15f | %-15f | %-11s | %-10s | %-10s | %-5s | %-10s |  %s' % (self.abstime, self.elength, hex(self.delta_time), self.etype, hex(self.event_id), mhxmet, hex(self.leng()),  mutils.hexstr(self.data, group= 2))
                return estr


from rmidi.dataset import config
from rmidi import mutils
from rmidi.MIDI import MIDI
from rmidi.absolutemidi import AbsoluteMidi

class reader():
    def __init__(self, dataset_path = config.PATH):
        if not dataset_path: raise AttributeError("set config.PATH to dataset path, or specify the path value to attribute 'dataset_path'")
        config.PATH = dataset_path

    def read(self, abs_midi = False, generator = True):
        midis = mutils.get_all_midis(config.PATH)
        res = []
        for fmid in midis:
            mid = MIDI.parse_midi(fmid)
            if abs_midi: mid = AbsoluteMidi.to_abs_midi(mid)
            yield mid
            res.append(mid)
        return res


from rmidi.MIDI import MIDI
import copy
from rmidi import mutils
class AbsoluteMidi(MIDI):
    def __init__(self, format_type=0, track_count=0, time_div=480, empty=False, filename=None, tempo = 500000):
        super().__init__(format_type=format_type, track_count=track_count, time_div=time_div, empty=empty, filename=filename) # no tempo change assumed glich
        self.tempo = tempo
        
    @classmethod
    def to_abs_midi(cls, midi = MIDI()):
        absmidi = copy.deepcopy(midi)
        # absmidi = midi
        abtime = 0
        tempo = 500000
        time_div = absmidi.time_div
        event_trk = [MIDI.Track(absmidi, empty= True) for _ in range(absmidi.track_count)]
        print(event_trk[0] is event_trk[1])
        for i, t in enumerate(absmidi.tracks):
            trk = event_trk[i]
            evts = []
            top = 0
            timekeeper = 0 #In milliseconds
            for e in t.trk_event:
                delt = e.delta_time
                if e.event_id == 0xFF : #Meta Event
                    if e.meta_event_type == 0x51: #set_tempo event
                        tempo = mutils.toint(e.data, 8)
                        e.abstime = timekeeper
                        e.elength = 0
                elif 0x7F <  e.event_id < 0xa0: #Note OFF / ON event
                    eventid = e.event_id >> 4
                    noteval = e.data[0]
                    velocity = e.data[1] 
                    timekeeper += (delt/time_div * tempo) / 60000
                    if eventid & 0xF == 9 and velocity != 0: #Note ON event
                        e.abstime = timekeeper
                        evts.append((e, top))

                    elif (eventid  & 0xF == 8) or velocity == 0: # Note OFF event
                        if not evts: raise ValueError("Note OFF Event Occured without the correspoint Note ON Event : ")
                        e_shot = evts.pop()
                        #trk.trk_event[e_shot[1]].elength = timekeeper - trk.trk_event[e_shot[1]].abstime
                        e_shot[0].elength = timekeeper - e_shot[0].abstime
                        continue #to avoid top ++, since Note OFF just update correspoind Note ON value
                else :
                    e.abstime = timekeeper
                    e.elength = 0

                trk.trk_event.append(e)
                top += 1
        # print(tempo)
        amidi = cls(absmidi.format_type, absmidi.track_count, absmidi.track_count, tempo = tempo)
        amidi.tracks = event_trk

        return amidi

    def __repr__(self):
        et = 0
        mstr = ""
        for t in self.tracks:
            # print(mstr)
            mstr += '______________________________________________________________________________________________________________________________ . . . \n'
            mstr += '| Absolute Time   |  Duration       |  Note  Time         |  Delta Time |  ETYPE     |   Event ID | META  | LENGTH     | DATA \n' #% (e.abstime, e.elength, hex(e.delta_time), e.etype, hex(e.event_id), mhxmet, hex(e.leng()),  mutils.hexstr(e.data, group= 2))
            mstr += '|______________________________________________________________________________________________________________________________ . . . \n'
            for e in t.trk_event:
                et += 1
                mhxmet = hex(e.meta_event_type) if e.meta_event_type else '0'
                # estr = " Abs Time : " + str(e.abstime) + "Duration : " + str(e.elength) + " |||>>>> Delta Time : " + hex(e.delta_time) + ", Etype : " + e.etype + ", Event ID : " + hex(e.event_id) + ", META : " + mhxmet + ", Length : " + hex(e.leng()) + ", Data --> " + mutils.hexstr(e.data, group= 2)
                estr = '| %-15f | %-15f | %-20s | %-11s | %-10s | %-10s | %-5s | %-10s |  %s' % (e.abstime, e.elength,mutils.nth_note(e.elength, self.tempo), hex(e.delta_time), e.etype, hex(e.event_id), mhxmet, hex(e.leng()),  mutils.hexstr(e.data, group= 2))
                mstr += (estr + "")
            mstr += '*******************************************************************************************************************\n\n'
        # le = len(mstr)
        # print(le, "  last : ", mstr[le - 1000: le])
        return mstr
        

    class AbsoluteTrack(MIDI.Track):
        def __init__(self, midi, empty=False):
            super().__init__(midi, empty=empty)

        class AbsoluteEvent(MIDI.Track.Event):
            def __init__(self, delta_time=0, etype=None, event_id=None, meta_event_type=None, length=0, data=bytearray()):
                    super().__init__(delta_time=delta_time, etype=etype, event_id=event_id, meta_event_type=meta_event_type, length=length, data=data)




import matplotlib.pyplot as plt
from rmidi.MIDI import MIDI, Constant
from rmidi.mutils import channel as mchannel
from rmidi import mutils
from rmidi.absolutemidi import AbsoluteMidi
from rmidi.constant import converter
import glob, os
import numpy
# import music21

class Analyser:
    def __init__(self, midi_file, **kwargs):
        self.__midi = MIDI.parse_midi(midi_file)
        
    def stats(self, names = False):
        stat_freq = {'channel' : {8: 0, 9: 0, 10: 0, 11: 0, 12: 0, 13: 0, 14: 0},
             'meta': {0: 0, 1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 32: 0, 33: 0, 47: 0, 81: 0, 84: 0, 88: 0, 89: 0, 127: 0}, 
             'sys': {240: 0, 247: 0}
             }
        for t in self.__midi.tracks:
            for e in t.trk_event:
                if e.is_channel_event():
                    stat_freq['channel'][mchannel(e.event_id)] += 1
                elif e.is_meta_event():
                    stat_freq['meta'][e.meta_event_type] += 1
                elif e.is_meta_event():
                    stat_freq['sys'][e.event_id] += 1
                else : raise AttributeError('invalid MIDI event or Event Id got is not supported by this version of rmidi ')
        if names:
            c_stat_freq = dict()
            evt_format = mutils.dictn(Constant.ch_event_format)
            c_stat_freq['channel'] = {evt_format[k][0] : v for k, v in stat_freq['channel'].items()}
            evt_format = mutils.dictn(Constant.meta_event_format)
            c_stat_freq['meta'] = {evt_format[k][0] : v for k, v in stat_freq['meta'].items()}            
            evt_format = mutils.dictn(Constant.sys_event_format)
            c_stat_freq['sys'] = {evt_format[k][0] : v for k, v in stat_freq['sys'].items()}
            return c_stat_freq
        return stat_freq 

    @classmethod
    def analyse_dataset(cls, folder_path, save = '.', st = 1, end = 100):
        files = mutils.get_all_midis(folder_path)
        zstats = {}
        ana = cls(files[0])
        res = ana.stats(True)
        try : 
            for k, v in res.items():
                zstats[k] = {}
                for ik , iv in v.items():
                    try:
                        zstats[k][ik] = iv
                    except Exception:
                        zstats[k][ik] = 0

            for f in files[st:end]:
                try:
                    ana = cls(f)
                    res = ana.stats(True)
                except Exception:
                    continue

                for k, v in res.items():
                    for ik , iv in v.items():
                        try:
                            zstats[k][ik] += iv
                        except Exception:
                            zstats[k][ik] = 0
                del ana
        except Exception:
            pass
        numpy.save(save + '\\_FST_' + str(st) + '_END_' + str(end) + '_NUMPYSTORE_DATSET_' + folder_path.split('\\')[-1], zstats)
        1 == 2

    def track_analysis(self, resolution = 32): #Instrument Based
        #probably may out of scale , strech or contraction may present at end
        amidi = AbsoluteMidi.to_abs_midi(self.__midi)
        # print(amidi)
        isec = 500 * (4 / resolution) #17.250  // In milliseconds
        
        # LEN = int((300 * 1000) // isec) # For 5 min (MAX) only
        LEN = 4800
        trackers = numpy.zeros(amidi.track_count)
        track_anas = converter.meta_event_format()
        res_set = {}
        for i, t in enumerate(amidi.tracks):
            instrument = t.get_event('instrument_name', depth = 1)
            try :
                ttempo = mutils.toint(t.get_event('set_tempo', depth = 1)[0].data, 8)
                tempo = ttempo
            except IndexError:
                pass
            res_arr =  numpy.zeros(LEN)
            res_set[instrument[0].data.decode("utf-8")  + " " + str(i)] = res_arr
            nit = 0 #numpy array iterator
            prev_oucr, dut = 0, 0
            # print(t.trk_event[-11 : -1])
            for e in t.trk_event:
                if e.is_note_on_off_event():
                    # print(e)
                    if prev_oucr == e.abstime: 
                        nit -= dut
                        # print(res_arr[nit: nit + dut])
                    dut = int(32 // mutils.nth_note(e.elength, tempo=tempo)) 
                    prev_oucr = e.abstime
                    res_arr[nit: nit + dut] += 1
                    # print(res_arr[nit: nit + dut])
                    nit += dut
            # print("--------------------------------------=============================----------------------------------")

        return res_set
            

if __name__ == "__main__":
    f = 'midis\\Believer_Imagine_Dragons'

    c = Analyser(f)

    g = c.stats()

    c == g


Mthd = (0x4d, 0x54, 0x68, 0x64)
Mtrk = (0x4d, 0x54, 0x72, 0x6b)
META_EVENT_END = (0x01, 0xff, 0x2f, 0x0)
MAX = 1048576
DTYPE = {'uint4': (0, 16), 'uint7': (0, 127), 'uint8' : (0, 256)}
CHANNEL_EVENT = 'CHANNEL'
META_EVENT = 'META'
SYS_EVENT = 'SYS'

SCALE = ['c-major', 'c#-major', 'd-major', 'd#-major', 'e-major', 'f-major', 'f#-major', 
            'g-major', 'g#-major', 'a-major', 'a#-major', 'b-major', 
            'c-minor', 'c#-minor', 'd-minor', 'd#-minor', 'e-minor', 'f-minor', 'f#-minor', 
            'g-minor', 'g#-minor', 'a-minor', 'a#-minor', 'b-minor'
            ]

SCALE_OFFSET =  {'c-major' : 0, 'c#-major' : 1, 'd-major' : 2, 'd#-major' : 3,'e-major' : 4 ,'f-major' : 5, 'f#-major' : 6, 
                    'g-major' : 7, 'g#-major' : 8, 'a-major' : 9, 'a#-major' : 10, 'b-major' : 11, 
                    'c-minor' : 3, 'c#-minor' : 4, 'd-minor' : 5, 'd#-minor' : 6, 'e-minor' : 7, 'f-minor' : 8, 'f#-minor' : 9, 
                    'g-minor' : 10, 'g#-minor' : 11, 'a-minor' : 0, 'a#-minor' : 1, 'b-minor' : 2 
                    }


"""
KEY_SIG ENCODING
Everything according to circle of fifth
no. of #'s  0 -> 7
no. of b's  8 -> 15

SCALE_KEY_SIG : Contains string to integer mapping info
SCALE_KEY_SIG_DEFAULT : Contains string to integer mapping info, no. of #'s prefer over b, # naming system in presence of both flats(b) & sharp(#)
SCALE_KEY_SIG_REV : Contains integer to string mapping info
"""

SCALE_KEY_SIG_REV = {0 : ('c-major', 'a-minor'), 1 : ('g-major', 'e-minor'), 2 : ('d-major', 'b-minor'), 3 : ('a-major', 'f#-minor'), 4 : ('e-major', 'c#-minor'), 5 : ('b-major', 'g#-minor'), 6 : ('f#-major', 'd#-minor'), 7 : ('d#-major', 'a#-minor'),
                    8 : ('g-major', 'e-minor'), 9 : ('f-major', 'd-minor'), 10 : ('a#-major', 'g-minor'), 11 : ('d#-major', 'c-minor'), 12 : ('g#-major', 'f-minor'), 13 : ('c#-major', 'a#-minor'), 14 : ('f#-major', 'd#-minor'), 15 : ('b-major', 'g#-minor')
                    }

SCALE_KEY_SIG =  {'c-major' : (0, 8), 'c#-major' : (7, 13) , 'd-major' : (2, 8), 'd#-major' : (0, 11),'e-major' : (4, 8) ,'f-major' : (0, 9), 'f#-major' : (6, 14), 
                    'g-major' : (1, 8), 'g#-major' : (0, 12), 'a-major' : (3, 8), 'a#-major' : (0, 10), 'b-major' : (5, 15), 
                    'c-minor' : (0, 11), 'c#-minor' : (4, 8), 'd-minor' : (0, 9), 'd#-minor' : (6, 14) , 'e-minor' : (1, 8), 'f-minor' : (0, 12), 'f#-minor' : (3, 8),  
                    'g-minor' : (0, 10), 'g#-minor' : (5, 15) , 'a-minor' : (0, 8), 'a#-minor' : (7, 13) , 'b-minor' : (2, 8) 
                    }


SCALE_KEY_SIG_DEFAULT =  {'c-major' : 0, 'c#-major' : 7, 'd-major' : 2, 'd#-major' : 11,'e-major' : 4 ,'f-major' : 9, 'f#-major' : 6, 
                    'g-major' : 1, 'g#-major' : 12, 'a-major' : 3, 'a#-major' : 10, 'b-major' : 5, 
                    'c-minor' : 11, 'c#-minor' : 4, 'd-minor' : 9, 'd#-minor' : 6 , 'e-minor' : 1, 'f-minor' : 12, 'f#-minor' : 3, 
                    'g-minor' : 10, 'g#-minor' : 5 , 'a-minor' : 0, 'a#-minor' : 7 , 'b-minor' : 2 
                    }

#format = (id, sub_event_name, param_len, params..., param_type)
ch_event_format = (
                    (0x8, "note_off", 2, "note_number", "velocity", "int"),
                    (0x9, "note_on", 2, "note_number", "velocity", "int"),
                    (0xA, "after_touch", 2, "note_number", "amount", "int"),
                    (0xB, "controller", 2, "controller_type", "value", "int"),                   
                    (0xC, "program_change", 1, "program_number", "int"),                   
                    (0xD, "channel_after_touch", 1, "amount", "int"),                   
                    (0xE, "pitch_bend", 2, "vlsb", "vmsb", "int")                  
                )
#format = (id, "sub_event_name", "param_len", params..., "param_type", "common_range, -1", //if -1, highest_allowed_val_of_param..., -1, "mask")
#If -1 in place of common range the following param len args are highesh values of parameters
#If all within mask positive else masked negative
meta_event_format = (
                        (0x00, "sequence_number", 2, "nmsb", "nlsb", "int", 255, 255),
                        (0x01, "text_event", -1, "text", "str"),
                        (0x02, "copyright_notice", -1, "text", "str"),
                        (0x03, "track_name", -1, "text", "str"),
                        (0x04, "instrument_name", -1, "text", "str"),
                        (0x05, "lyrics", -1, "text", "str"),
                        (0x06, "marker", -1, "text", "str"),
                        (0x07, "cue_point", -1, "text", "str"),
                        (0x20, "midi_ch_prefix", 1, "channel", "int", 15),
                        (0x21, "midi_port", 1, "port_no", "int", 15),
                        (0x2F, "end_of_track", 0, None, "None"),
                        (0x51, "set_tempo", 3, "musec_per_quat_note", "musec_per_quat_note", "musec_per_quat_note", "int", 127, 127, 127),
                        (0x54, "smpte_offset", 5, "hr", "min", "sec", "fr", "subfr", "int", 24, 60, 60, 30, 100),
                        (0x58, "time_sig", 4, "numer", "denom", "metro", "32nds", "int", 255, 255, 255, 255),
                        (0x59, "key_sig", 2, "key", "scale", "int" , 15, 1 ),
                        (0x7F, "sequence_specifier", -1, "data", "any")
                    )
sys_event_format = (
                        (0xF0, "normal_sys_event", -1, "data", "int", 128),
                        (0xF7, "authorization_sys_event", -1, "data", "int", 256),
                        
                    )

#controller format = (controller_value/id , "controller_type", "len_of_simmilar_extension")
controller = (
                (0x00, "blank_select", 1),
                (0x01, "modulation", 1),
                (0x02, "breath_controller", 1),
                (0x04, "foot_controller", 1),
                (0x05, "portamento_time", 1),
                (0x06, "data_entry", 1),#May be msb
                (0x07, "main_volume", 1),
                (0x08, "balance", 1),
                (0x0A, "pan", 1),
                (0x0B, "expression_controller", 1),
                (0x0C, "effect_control", 2),
                # (0x0D, "effect_control_2", 0),
                (0x10, "general_purpose_controller", 4),#, 0x50, 4),
                (0x20, "LSB_for_controller", 32),
                (0x40, "damper_pedal", 1),
                (0x41, "portamento", 1),
                (0x42, "sostenuto", 1),
                (0x43, "soft_pedal", 1),
                (0x44, "legato_footswitch", 1),
                (0x45, "hold", 1),
                # (0x46, "damper_pedal", 1),
                (0x46, "sound_controller", 10),
                # (0x40, "damper_pedal", 1),
                (0x50, "post_general_purpose_controller", 4),
                (0x54, "portamento_control", 1),
                (0x5B, "effects_depth", 5),
                (0x60, "data_increment", 1),
                (0x61, "data_decrement", 1),
                (0x62, "nonreg_param_num_lsb", 1),
                (0x63, "nonreg_param_num_msb", 1),
                (0x64, "reg_param_num_lsb", 1),
                (0x65, "reg_param_num_msb", 1),
                (0x79, "mode_messages", 7),
            )


ch_event_format = {
    8: {'id': 8,
        'length': 2,
        'mask': 127,
        'name': 'note_off',
        'params': ('note_number', 'velocity')},
    9: {'id': 9,
        'length': 2,
        'mask': 127,
        'name': 'note_on',
        'params': ('note_number', 'velocity')},
    10: {'id': 10,
        'length': 2,
        'mask': 127,
        'name': 'after_touch',
        'params': ('note_number', 'amount')},
    11: {0: {'cntrl_id': 0, 'cntrl_name': 'blank_select'},
        1: {'cntrl_id': 1, 'cntrl_name': 'modulation'},
        2: {'cntrl_id': 2, 'cntrl_name': 'breath_controller'},
        4: {'cntrl_id': 4, 'cntrl_name': 'foot_controller'},
        5: {'cntrl_id': 5, 'cntrl_name': 'portamento_time'},
        6: {'cntrl_id': 6, 'cntrl_name': 'data_entry'},
        7: {'cntrl_id': 7, 'cntrl_name': 'main_volume'},
        8: {'cntrl_id': 8, 'cntrl_name': 'balance'},
        10: {'cntrl_id': 10, 'cntrl_name': 'pan'},
        11: {'cntrl_id': 11, 'cntrl_name': 'expression_controller'},
        12: {'cntrl_id': 12, 'cntrl_name': 'effect_control_0'},
        13: {'cntrl_id': 13, 'cntrl_name': 'effect_control_1'},
        16: {'cntrl_id': 16, 'cntrl_name': 'general_purpose_controller_0'},
        17: {'cntrl_id': 17, 'cntrl_name': 'general_purpose_controller_1'},
        18: {'cntrl_id': 18, 'cntrl_name': 'general_purpose_controller_2'},
        19: {'cntrl_id': 19, 'cntrl_name': 'general_purpose_controller_3'},
        32: {'cntrl_id': 32, 'cntrl_name': 'LSB_for_controller_0'},
        33: {'cntrl_id': 33, 'cntrl_name': 'LSB_for_controller_1'},
        34: {'cntrl_id': 34, 'cntrl_name': 'LSB_for_controller_2'},
        35: {'cntrl_id': 35, 'cntrl_name': 'LSB_for_controller_3'},
        36: {'cntrl_id': 36, 'cntrl_name': 'LSB_for_controller_4'},
        37: {'cntrl_id': 37, 'cntrl_name': 'LSB_for_controller_5'},
        38: {'cntrl_id': 38, 'cntrl_name': 'LSB_for_controller_6'},
        39: {'cntrl_id': 39, 'cntrl_name': 'LSB_for_controller_7'},
        40: {'cntrl_id': 40, 'cntrl_name': 'LSB_for_controller_8'},
        41: {'cntrl_id': 41, 'cntrl_name': 'LSB_for_controller_9'},
        42: {'cntrl_id': 42, 'cntrl_name': 'LSB_for_controller_10'},
        43: {'cntrl_id': 43, 'cntrl_name': 'LSB_for_controller_11'},
        44: {'cntrl_id': 44, 'cntrl_name': 'LSB_for_controller_12'},
        45: {'cntrl_id': 45, 'cntrl_name': 'LSB_for_controller_13'},
        46: {'cntrl_id': 46, 'cntrl_name': 'LSB_for_controller_14'},
        47: {'cntrl_id': 47, 'cntrl_name': 'LSB_for_controller_15'},
        48: {'cntrl_id': 48, 'cntrl_name': 'LSB_for_controller_16'},
        49: {'cntrl_id': 49, 'cntrl_name': 'LSB_for_controller_17'},
        50: {'cntrl_id': 50, 'cntrl_name': 'LSB_for_controller_18'},
        51: {'cntrl_id': 51, 'cntrl_name': 'LSB_for_controller_19'},
        52: {'cntrl_id': 52, 'cntrl_name': 'LSB_for_controller_20'},
        53: {'cntrl_id': 53, 'cntrl_name': 'LSB_for_controller_21'},
        54: {'cntrl_id': 54, 'cntrl_name': 'LSB_for_controller_22'},
        55: {'cntrl_id': 55, 'cntrl_name': 'LSB_for_controller_23'},
        56: {'cntrl_id': 56, 'cntrl_name': 'LSB_for_controller_24'},
        57: {'cntrl_id': 57, 'cntrl_name': 'LSB_for_controller_25'},
        58: {'cntrl_id': 58, 'cntrl_name': 'LSB_for_controller_26'},
        59: {'cntrl_id': 59, 'cntrl_name': 'LSB_for_controller_27'},
        60: {'cntrl_id': 60, 'cntrl_name': 'LSB_for_controller_28'},
        61: {'cntrl_id': 61, 'cntrl_name': 'LSB_for_controller_29'},
        62: {'cntrl_id': 62, 'cntrl_name': 'LSB_for_controller_30'},
        63: {'cntrl_id': 63, 'cntrl_name': 'LSB_for_controller_31'},
        64: {'cntrl_id': 64, 'cntrl_name': 'damper_pedal'},
        65: {'cntrl_id': 65, 'cntrl_name': 'portamento'},
        66: {'cntrl_id': 66, 'cntrl_name': 'sostenuto'},
        67: {'cntrl_id': 67, 'cntrl_name': 'soft_pedal'},
        68: {'cntrl_id': 68, 'cntrl_name': 'legato_footswitch'},
        69: {'cntrl_id': 69, 'cntrl_name': 'hold'},
        70: {'cntrl_id': 70, 'cntrl_name': 'sound_controller_0'},
        71: {'cntrl_id': 71, 'cntrl_name': 'sound_controller_1'},
        72: {'cntrl_id': 72, 'cntrl_name': 'sound_controller_2'},
        73: {'cntrl_id': 73, 'cntrl_name': 'sound_controller_3'},
        74: {'cntrl_id': 74, 'cntrl_name': 'sound_controller_4'},
        75: {'cntrl_id': 75, 'cntrl_name': 'sound_controller_5'},
        76: {'cntrl_id': 76, 'cntrl_name': 'sound_controller_6'},
        77: {'cntrl_id': 77, 'cntrl_name': 'sound_controller_7'},
        78: {'cntrl_id': 78, 'cntrl_name': 'sound_controller_8'},
        79: {'cntrl_id': 79, 'cntrl_name': 'sound_controller_9'},
        80: {'cntrl_id': 80, 'cntrl_name': 'post_general_purpose_controller_0'},
        81: {'cntrl_id': 81, 'cntrl_name': 'post_general_purpose_controller_1'},
        82: {'cntrl_id': 82, 'cntrl_name': 'post_general_purpose_controller_2'},
        83: {'cntrl_id': 83, 'cntrl_name': 'post_general_purpose_controller_3'},
        84: {'cntrl_id': 84, 'cntrl_name': 'portamento_control'},
        91: {'cntrl_id': 91, 'cntrl_name': 'effects_depth_0'},
        92: {'cntrl_id': 92, 'cntrl_name': 'effects_depth_1'},
        93: {'cntrl_id': 93, 'cntrl_name': 'effects_depth_2'},
        94: {'cntrl_id': 94, 'cntrl_name': 'effects_depth_3'},
        95: {'cntrl_id': 95, 'cntrl_name': 'effects_depth_4'},
        96: {'cntrl_id': 96, 'cntrl_name': 'data_increment'},
        97: {'cntrl_id': 97, 'cntrl_name': 'data_decrement'},
        98: {'cntrl_id': 98, 'cntrl_name': 'nonreg_param_num_lsb'},
        99: {'cntrl_id': 99, 'cntrl_name': 'nonreg_param_num_msb'},
        100: {'cntrl_id': 100, 'cntrl_name': 'reg_param_num_lsb'},
        101: {'cntrl_id': 101, 'cntrl_name': 'reg_param_num_msb'},
        121: {'cntrl_id': 121, 'cntrl_name': 'mode_messages_0'},
        122: {'cntrl_id': 122, 'cntrl_name': 'mode_messages_1'},
        123: {'cntrl_id': 123, 'cntrl_name': 'mode_messages_2'},
        124: {'cntrl_id': 124, 'cntrl_name': 'mode_messages_3'},
        125: {'cntrl_id': 125, 'cntrl_name': 'mode_messages_4'},
        126: {'cntrl_id': 126, 'cntrl_name': 'mode_messages_5'},
        127: {'cntrl_id': 127, 'cntrl_name': 'mode_messages_6'},
        'id': 11,
        'length': 2,
        'mask': 127,
        'params': ['controller_type', 'value']},
    12: {'id': 12,
        'length': 1,
        'mask': 127,
        'name': 'program_change',
        'params': ('program_number',)},
    13: {'id': 13,
        'length': 1,
        'mask': 127,
        'name': 'channel_after_touch',
        'params': ('amount',)},
    14: {'id': 14,
        'length': 2,
        'mask': 127,
        'name': 'pitch_bend',
        'params': ('vlsb', 'vmsb')
        }
}



meta_event_format_dict = {
        0xFF: {
            0: {'dtype': 'int',
                'length': 2,
                'mask': (255, 255),
                'params': ('nmsb', 'nlsb'),
                'type_id': 0,
                'type_name': 'sequence_number'},
            1: {'dtype': 'str',
                'length': -1,
                'mask': 127,
                'params': 'text',
                'type_id': 1,
                'type_name': 'text_event'},
            2: {'dtype': 'str',
                'length': -1,
                'mask': 127,
                'params': 'text',
                'type_id': 2,
                'type_name': 'copyright_notice'},
            3: {'dtype': 'str',
                'length': -1,
                'mask': 127,
                'params': 'text',
                'type_id': 3,
                'type_name': 'track_name'},
            4: {'dtype': 'str',
                'length': -1,
                'mask': 127,
                'params': 'text',
                'type_id': 4,
                'type_name': 'instrument_name'},
            5: {'dtype': 'str',
                'length': -1,
                'mask': 127,
                'params': 'text',
                'type_id': 5,
                'type_name': 'lyrics'},
            6: {'dtype': 'str',
                'length': -1,
                'mask': 127,
                'params': 'text',
                'type_id': 6,
                'type_name': 'marker'},
            7: {'dtype': 'str',
                'length': -1,
                'mask': 127,
                'params': 'text',
                'type_id': 7,
                'type_name': 'cue_point'},
            32: {'dtype': 'int',
                    'length': 1,
                    'mask': (15,),
                    'params': ('channel',),
                    'type_id': 32,
                    'type_name': 'midi_ch_prefix'},
            33: {'dtype': 'int',
                    'length': 1,
                    'mask': (15,),
                    'params': ('port_no',),
                    'type_id': 33,
                    'type_name': 'midi_port'},
            47: {'dtype': 'None',
                    'length': 0,
                    'mask': 127,
                    'params': None,
                    'type_id': 47,
                    'type_name': 'end_of_track'},
            81: {'dtype': 'int',
                    'length': 3,
                    'mask': (127, 127, 127),
                    'params': ('musec_per_quat_note',
                            'musec_per_quat_note',
                            'musec_per_quat_note'),
                    'type_id': 81,
                    'type_name': 'set_tempo'},
            84: {'dtype': 'int',
                    'length': 5,
                    'mask': (24, 60, 60, 30, 100),
                    'params': ('hr', 'min', 'sec', 'fr', 'subfr'),
                    'type_id': 84,
                    'type_name': 'smpte_offset'},
            88: {'dtype': 'int',
                    'length': 4,
                    'mask': (255, 255, 255, 255),
                    'type_id': 88,
                    'type_name': 'time_sig'},
            0x59: {'dtype': 'int',
                    'length': 2,
                    'mask': (15, 1),
                    'params': ('key', 'scale'),
                    'type_id': 89,
                    'type_name': 'key_sig'},
            127: {'dtype': 'any',
                    'length': -1,
                    'mask': 127,
                    'params': 'text',
                    'type_id': 127,
                    'type_name': 'sequence_specifier'},
            'id': 255,
            'name': 'meta'
            }
}



sys_event_format_dict = {
    240: {
        "id": 240,
        "name": "normal_sys_event",
        "length": -1,
        "params": "data",
        "dtype": "int",
        "mask": 128
    },
    247: {
        "id": 247,
        "name": "authorization_sys_event",
        "length": -1,
        "params": "data",
        "dtype": "int",
        "mask": 256
    }
}


# X = {
#     "0xFF": {
#         "id": 255,
#         "name": "meta",
#         "0x00": {
#             "type_id": 0,
#             "type_name": "sequence_number",
#             "length": 2,
#             "params": [
#                 "nmsb",
#                 "nlsb"
#             ],
#             "dtype": "int",
#             "mask": [
#                 255,
#                 255
#             ],
#             "default" : [0, 0]
#         },
#         "0x01": {
#             "type_id": 1,
#             "type_name": "text_event",
#             "length": -1,
#             "params": "text",
#             "dtype": "str",
#             "mask": 127,
#             "default" : "Enter the Text"
#         },
#         "0x02": {
#             "type_id": 2,
#             "type_name": "copyright_notice",
#             "length": -1,
#             "params": "text",
#             "dtype": "str",
#             "mask": 127,
#             "default" : "No Copyright"
#         },
#         "0x03": {
#             "type_id": 3,
#             "type_name": "track_name",
#             "length": -1,
#             "params": "text",
#             "dtype": "str",
#             "mask": 127,
#             "default" : "Track 0"
#         },
#         "0x04": {
#             "type_id": 4,
#             "type_name": "instrument_name",
#             "length": -1,
#             "params": "text",
#             "dtype": "str",
#             "mask": 127,
#             "default" : "Piano"
#         },
#         "0x05": {
#             "type_id": 5,
#             "type_name": "lyrics",
#             "length": -1,
#             "params": "text",
#             "dtype": "str",
#             "mask": 127,
#             "default" : "LYRICS"
#         },
#         "0x06": {
#             "type_id": 6,
#             "type_name": "marker",
#             "length": -1,
#             "params": "text",
#             "dtype": "str",
#             "mask": 127,
#             "default" : "*"
#         },
#         "0x07": {
#             "type_id": 7,
#             "type_name": "cue_point",
#             "length": -1,
#             "params": "text",
#             "dtype": "str",
#             "mask": 127,
#             "default" : "-"
#         },
#         "0x20": {
#             "type_id": 32,
#             "type_name": "midi_ch_prefix",
#             "length": 1,
#             "params": [
#                 "channel"
#             ],
#             "dtype": "int",
#             "mask": [
#                 15
#             ],
#             "default" : [0]
#         },
#         "0x21": {
#             "type_id": 33,
#             "type_name": "midi_port",
#             "length": 1,
#             "params": [
#                 "port_no"
#             ],
#             "dtype": "int",
#             "mask": [
#                 15
#             ],
#             "default" : [0]
#         },
#         "0x2F": {
#             "type_id": 47,
#             "type_name": "end_of_track",
#             "length": 0,
#             "params": None,
#             "dtype": "None",
#             "mask": 127,
#             "default" : None
#         },
#         "0x51": {
#             "type_id": 81,
#             "type_name": "set_tempo",
#             "length": 3,
#             "params": [
#                 "musec_per_quat_note",
#                 "musec_per_quat_note",
#                 "musec_per_quat_note"
#             ],
#             "dtype": "int",
#             "mask": [
#                 127,
#                 127,
#                 127
#             ],
#             "default" : [0, 0, 0]
#         },
#         "0x54": {
#             "type_id": 84,
#             "type_name": "smpte_offset",
#             "length": 5,
#             "params": [
#                 "hr",
#                 "min",
#                 "sec",
#                 "fr",
#                 "subfr"
#             ],
#             "dtype": "int",
#             "mask": [
#                 24,
#                 60,
#                 60,
#                 30,
#                 100
#             ],
#             "default" : [0, 0 , 0, 0, 0]
#         },
#         "0x58": {
#             "type_id": 88,
#             "type_name": "time_sig",
#             "length": 4,
#             "params": [
#                 "numer",
#                 "denom",
#                 "metro",
#                 "32nds"
#             ],
#             "dtype": "int",
#             "mask": [
#                 255,
#                 255,
#                 255,
#                 255
#             ],
#             "default" : [0, 0, 0, 0]
#         },
#         "0x59": {
#             "type_id": 89,
#             "type_name": "key_sig",
#             "length": 2,
#             "params": [
#                 "key",
#                 "scale"
#             ],
#             "dtype": "int",
#             "mask": [
#                 15,
#                 1
#             ],
#             "default" : [0, 1]
#         },
#         "0x7F": {
#             "type_id": 127,
#             "type_name": "sequence_specifier",
#             "length": -1,
#             "params": "text",
#             "dtype": "any",
#             "mask": 127,
#             "default" : [0]
#         }
#     }
# }

X = {
        255: {
            0: {'dtype': 'int',
                'length': 2,
                'mask': (255, 255),
                'params': ('nmsb', 'nlsb'),
                'type_id': 0,
                'type_name': 'sequence_number',
                "default" : [0, 0]
                },
            1: {'dtype': 'str',
                'length': -1,
                'mask': 127,
                'params': 'text',
                'type_id': 1,
                'type_name': 'text_event',
                "default" : "Enter the Text"
                },
            2: {'dtype': 'str',
                'length': -1,
                'mask': 127,
                'params': 'text',
                'type_id': 2,
                'type_name': 'copyright_notice',
                "default" : "No Copyright"
                },
            3: {'dtype': 'str',
                'length': -1,
                'mask': 127,
                'params': 'text',
                'type_id': 3,
                'type_name': 'track_name',
                "default" : "Track 0"
                },
            4: {'dtype': 'str',
                'length': -1,
                'mask': 127,
                'params': 'text',
                'type_id': 4,
                'type_name': 'instrument_name',
                "default" : "Piano"
                },
            5: {'dtype': 'str',
                'length': -1,
                'mask': 127,
                'params': 'text',
                'type_id': 5,
                'type_name': 'lyrics',
                "default" : "LYRICS"
                },
            6: {'dtype': 'str',
                'length': -1,
                'mask': 127,
                'params': 'text',
                'type_id': 6,
                'type_name': 'marker',
                "default" : "*"
                },
            7: {'dtype': 'str',
                'length': -1,
                'mask': 127,
                'params': 'text',
                'type_id': 7,
                'type_name': 'cue_point',
                "default" : "-"
                },
            32: {'dtype': 'int',
                    'length': 1,
                    'mask': (15,),
                    'params': ('channel',),
                    'type_id': 32,
                    'type_name': 'midi_ch_prefix',
                    "default" : [0]
                    },
            33: {'dtype': 'int',
                    'length': 1,
                    'mask': (15,),
                    'params': ('port_no',),
                    'type_id': 33,
                    'type_name': 'midi_port',
                    "default" : [0]
                    },
            47: {'dtype': 'None',
                    'length': 0,
                    'mask': 127,
                    'params': None,
                    'type_id': 47,
                    'type_name': 'end_of_track',
                    "default" : None
                    },
            81: {'dtype': 'int',
                    'length': 3,
                    'mask': (127, 127, 127),
                    'params': ('musec_per_quat_note',
                            'musec_per_quat_note',
                            'musec_per_quat_note'),
                    'type_id': 81,
                    'type_name': 'set_tempo',
                    "default" : [0, 0, 0]
                    },
            84: {'dtype': 'int',
                    'length': 5,
                    'mask': (24, 60, 60, 30, 100),
                    'params': ('hr', 'min', 'sec', 'fr', 'subfr'),
                    'type_id': 84,
                    'type_name': 'smpte_offset',
                    "default" : [0, 0, 0, 0, 0]
                    },
            88: {'dtype': 'int',
                    'length': 4,
                    'mask': (255, 255, 255, 255),
                    'type_id': 88,
                    'type_name': 'time_sig',
                    "default" : [0, 0, 0, 0]
                    },
            89: {'dtype': 'int',
                    'length': 2,
                    'mask': (15, 1),
                    'params': ('key', 'scale'),
                    'type_id': 89,
                    'type_name': 'key_sig',
                    "default" : [0, 0]
                    },
            127: {'dtype': 'any',
                    'length': -1,
                    'mask': 127,
                    'params': 'text',
                    'type_id': 127,
                    'type_name': 'sequence_specifier',
                    "default" : [0,]
                    },
            'id': 255,
            'name': 'meta'
            }
        }


import math, re
import numpy
import hashlib as ha
import os, glob

hash_ =  ha.md5()

WRAP_DATA = 0x7f
WRAP_BITS = 7

def hex2(val):
    return "0x{:02x}".format(val)

@DeprecationWarning
def hexn(val, n):
    return "0x{:0{}x}".format(val, n)

def ch_event_id(id, no) :
    return ((id & 0xf) << 4) | (no & 0xf)

def channel(evt_id):
    return (evt_id >> 4) & 0xf
    
def numin(num, startb, off):#off length of bits that want
    return (num >> startb) & ((1 << off) - 1)

def meta_event_type(typ):
    return typ & 0xff

def invert(num, bits = 3, twos = False):
    # if bits == 3: return ~num & 7
    twos = 1 if twos else 0
    MASK = (1 << bits) - 1
    return ((~num & MASK) + twos)  & MASK  

def magnitude(num, bits):
    """Only use for signed integer
    
    Arguments:
        num {number} -- mag(num)
        bits {number} -- bits used to encode
    """
    MASK = (1 << bits - 1) - 1
    if num > MASK: return ((~num & MASK) + 1)
    return num & MASK
    

def to_var_length(k):
    if k > 127:
        leng = length(k) // WRAP_BITS + 1
        var = bytearray(leng)
        var[-1] = k & WRAP_DATA
        k >>= WRAP_BITS
        for i in range(leng - 2, -1, -1):
            var[i] = (k & WRAP_DATA) | (1 << WRAP_BITS )
            k >>= WRAP_BITS
        return var
    else:
        return bytearray((k,))     

def to_fix_length(k, leng, bits):
    fix = bytearray(leng)
    wrapper = (1 << bits) - 1
    if k > 255 and k > -1:
        fix[-1] = k & wrapper
        k >>= bits
        for i in range(leng - 2, -1, -1):
            # fix[i] = (k & wrapper) + wrapper + 1
            fix[i] = k & wrapper
            k >>= bits
    elif k > -1 :
        fix[-1] = k & 0xff
    return fix


def vartoint(varray:bytearray()):
    return toint(varray, 7)

def toint(a : bytearray(), bits = 8, mode = 'BG'):
    WRAPPER = (1 << bits) - 1
    num, itr, ind, s, le = 0x00, 0, 0, 1, len(a)
    if mode == 'LL': #set loop from end for little indian  
        s, ind= -1, le - 1
    while itr < le and ind < le:
        num = (num << bits) | (a[ind] & WRAPPER)
        ind += s
        itr += 1
    return num


def up(n_b, base = 2):
    return base ** int(math.log2(n_b) / math.log2(base) + 1)
    
def split(n, t, block_size = None): #Splits 'n' integer in t integer of bits bit(n)/t
    n_b = length(n)
    sbit = up(n_b) // t if not block_size else block_size
    WRAPPER = 2 ** sbit - 1
    k = n
    li = []
    while k:
        li.append(k & WRAPPER)
        k >>= sbit
    le = t - len(li)
    for _ in range(le):
        li.append(0)
    li.reverse()
    return li


def merge(a, b, *nums, block_size = 16):#Merge the numbers in order
    WRAPPER = (1 << block_size) - 1
    num = ((a & WRAPPER) << block_size) + (b & WRAPPER)
    for v in nums:
        num = (num << block_size) + (v & WRAPPER)
    return num 

def length(k):
    """Finds length of k in bits
    
    Arguments:
        k {int} -- Integer number
    """
    return int(math.log2(k) + 1)

def match(whole:bytearray, pattern= bytearray):
    return re.search(pattern, whole)

def find_location(text, listt):
    try:
        return next((i, j) 
            for i, t in enumerate(listt)
            for j, v in enumerate(t)
            if v == text)
    except StopIteration:
        return None

# @param num byte array to be converted
# @param type type : big-endian or small-endian
# @param group no. of bytes in group
# @param length length of line(in bytes)
# @return
def hexstr(bnum: bytearray, leng = 0, group = 0, numlen = 2, ftype = 0):
    if ftype == 1: bnum = bnum.reverse()
    if group == 0: group = len(bnum)
    if leng == 0: leng = len(bnum)    
    st = ''
    for i in range(len(bnum)):
        x = hex(bnum[i])[2:]
        if len(x) != numlen: x = "0" * (numlen - len(x)) + x[:2]
        st += ('0x' + x + ' ')
        if i % group == 0: st += ' '
        if (i + 1) % leng == 0: st += '\n'
    return st + '\n' if st == '' else st

def dtime(delta_time, time_div):
    if not delta_time: return 0
    return int((time_div * 4) // delta_time)


def file_hash(f, hexst = False):
    with open(f, "rb") as fi:
        cont = fi.read()
        return ha.md5(cont).hexdigest()
    return hash_.hexdigest()

def midi_to_note(noteval):
    if not numpy.isscalar(noteval): return [midi_to_note(v) for v in noteval]

    if not 0 <= noteval < 128: raise ValueError('Noteval not in range : {}'.format(noteval))
    
    mod = ['c', 'c#', 'd', 'd#', 'e', 'f', 'f#', 'g', 'g#', 'a', 'a#', 'b']
    
    note = mod[noteval % 12]
    octave = noteval // 12 - 1
    return '{}{}'.format(note, octave)

@DeprecationWarning
def note_to_midi(note): #Not implemented
    val_notes = {'c' : 0, 'c#' : 1, 'd' : 2, 'd#': 3, 'e': 4, 'f': 5, 'f#': 6, 'g': 7, 'g#': 8, 'a': 9, 'a#': 10, 'b': 11}
    if not numpy.isscalar(note): return [note_to_midi(v) for v in note]
    
    if note[:2].lower() not in val_notes: raise ValueError('Invalid Note String {}'.format(note))

    note, octave = note[:2], note[2:]


def dictn(ndlist):
    return {l[0] : l[1:]  for l in ndlist}

def get_all_midis(folder_path):
    # os.chdir(folder_path)
    f1, f2 = [], []
    for root, dirs, files in os.walk(folder_path, topdown=False):
        # for name in files:
        #     print(os.path.join(root, name))
        for name in dirs:
            print(os.path.join(root, name))
            p = os.path.join(root, name)
            f1 += glob.glob(p + '\*.mid')
            f2 += glob.glob(p + '\*.midi')
    return f1 + f2

def quater_note_to_millis(tempo):
    return tempo / 60000

def nth_note(duration, tempo):
    th2 = quater_note_to_millis(tempo) / 8
    wh_note = th2 * 32
    to_note = duration / th2
    rt = int(numpy.round(to_note))
    # rt_str = '{0:05b}'.format(rt)
    fin = 32 / rt if rt != 0 else 0
    return fin



def find_in_nested_dict(dictn, value, depth = 0):
    """ return dict containg the value in nested dict
    
    Arguments:
        dictn {dict} -- nested dictionart
        value {int} -- value to search
    
    Keyword Arguments:
        depth {int} -- depth value in nested dictionary (default: {0})
    
    Raises:
        ValueError: if value is not present in dict
    
    Returns:
        [dict] -- dict containing the value
    """
    for k, v in dictn:
        if v == value:
            return dictn
        elif hasattr(v, 'items'): # indicates if dictionary
            return find_in_nested_dict(v, value, depth - 1)
        else : raise ValueError("Value not found in the nested dictionary")

def reduce_dim(n, factor = 8):
    T = len(n) // factor
    res = numpy.zeros(T)
    
    for i in range(T):
        res[i] = sum(n[i : i + factor]) / factor

    return res


from rmidi.dataset import config
from rmidi import mutils
from rmidi.MIDI import MIDI
from rmidi.absolutemidi import AbsoluteMidi
from rmidi.constant import converter
import numpy

class PianoRoll:
    def __init__(self, midi, print_threshold = 25):
        self.midi = midi
        self.print_threshold = print_threshold
        self.roll = None
        pass

    def pianoroll(self):
        if not isinstance(self.midi, AbsoluteMidi):
            self.midi = AbsoluteMidi.to_abs_midi(self.midi)
        LEN = 4800
        trackers = numpy.zeros(self.midi.track_count)
        track_anas = converter.meta_event_format()
        res_set = numpy.zeros((self.midi.track_count, 128, LEN))
        for i, t in enumerate(self.midi.tracks):
            instrument = t.get_event('instrument_name', depth = 1)
            try :
                ttempo = mutils.toint(t.get_event('set_tempo', depth = 1)[0].data, 8)
                tempo = ttempo
            except IndexError:
                pass
            nit = 0 #numpy array iterator
            prev_oucr, dut = 0, 0
            # print(t.trk_event[-11 : -1])
            for e in t.trk_event:
                if e.is_note_on_off_event():
                    # print(e)
                    if prev_oucr == e.abstime: 
                        nit -= dut
                    noteval = e.data[0]
                    dut = int(32 // mutils.nth_note(e.elength, tempo=tempo)) 
                    prev_oucr = e.abstime
                    res_set[i][noteval][nit: nit + dut] += 1
                    nit += dut
        self.roll = res_set
        return res_set

    def to_str(self, res = None):
        p_str = ""
        tracks, notes, intervals = numpy.shape(self.roll)
        for t in range(tracks):
            for n in range(notes):
                p_str += mutils.midi_to_note(n) + " : "
                for i in  range(self.print_threshold):
                    p_str += ( str(self.roll[t][n][i]) + " -- " )

                p_str += "\n"

            p_str += "************************ NEW TRACK ***************************************************** NEW TRACK *********************************"
        return p_str
        
    def __repr__(self):
        p_str = ""
        tracks, notes, intervals = numpy.shape(self.roll)
        for t in range(tracks):
            for n in range(notes):
                p_str += mutils.midi_to_note(n) + " : "
                for i in  range(self.print_threshold):
                    p_str += ( str(self.roll[t][n][i]) + " -- " )

                p_str += "\n"

            p_str += "************************ NEW TRACK ***************************************************** NEW TRACK *********************************"
        return p_str
            
            

         


# X = {
#     "8": {
#         "id": 8,
#         "name": "note_off",
#         "length": 2,
#         "params": [
#             "note_number",
#             "velocity"
#         ],
#         "mask": 127
#     },
#     "9": {
#         "id": 9,
#         "name": "note_on",
#         "length": 2,
#         "params": [
#             "note_number",
#             "velocity"
#         ],
#         "mask": 127
#     },
#     "10": {
#         "id": 10,
#         "name": "after_touch",
#         "length": 2,
#         "params": [
#             "note_number",
#             "amount"
#         ],
#         "mask": 127
#     },
#     "11": {
#         "id": 11,
#         "name": "controller",
#         "length": 2,
#         "params": [
#             "controller_type",
#             "value"
#         ],
#         "mask": 127,
#         "0": {
#             "cntrl_id": 0,
#             "cntrl_name": "blank_select"
#         },
#         "1": {
#             "cntrl_id": 1,
#             "cntrl_name": "modulation"
#         },
#         "2": {
#             "cntrl_id": 2,
#             "cntrl_name": "breath_controller"
#         },
#         "4": {
#             "cntrl_id": 4,
#             "cntrl_name": "foot_controller"
#         },
#         "5": {
#             "cntrl_id": 5,
#             "cntrl_name": "portamento_time"
#         },
#         "6": {
#             "cntrl_id": 6,
#             "cntrl_name": "data_entry"
#         },
#         "7": {
#             "cntrl_id": 7,
#             "cntrl_name": "main_volume"
#         },
#         "8": {
#             "cntrl_id": 8,
#             "cntrl_name": "balance"
#         },
#         "10": {
#             "cntrl_id": 10,
#             "cntrl_name": "pan"
#         },
#         "11": {
#             "cntrl_id": 11,
#             "cntrl_name": "expression_controller"
#         },
#         "12": {
#             "cntrl_id": 12,
#             "cntrl_name": "effect_control_0"
#         },
#         "13": {
#             "cntrl_id": 13,
#             "cntrl_name": "effect_control_1"
#         },
#         "16": {
#             "cntrl_id": 16,
#             "cntrl_name": "general_purpose_controller_0"
#         },
#         "17": {
#             "cntrl_id": 17,
#             "cntrl_name": "general_purpose_controller_1"
#         },
#         "18": {
#             "cntrl_id": 18,
#             "cntrl_name": "general_purpose_controller_2"
#         },
#         "19": {
#             "cntrl_id": 19,
#             "cntrl_name": "general_purpose_controller_3"
#         },
#         "32": {
#             "cntrl_id": 32,
#             "cntrl_name": "LSB_for_controller_0"
#         },
#         "33": {
#             "cntrl_id": 33,
#             "cntrl_name": "LSB_for_controller_1"
#         },
#         "34": {
#             "cntrl_id": 34,
#             "cntrl_name": "LSB_for_controller_2"
#         },
#         "35": {
#             "cntrl_id": 35,
#             "cntrl_name": "LSB_for_controller_3"
#         },
#         "36": {
#             "cntrl_id": 36,
#             "cntrl_name": "LSB_for_controller_4"
#         },
#         "37": {
#             "cntrl_id": 37,
#             "cntrl_name": "LSB_for_controller_5"
#         },
#         "38": {
#             "cntrl_id": 38,
#             "cntrl_name": "LSB_for_controller_6"
#         },
#         "39": {
#             "cntrl_id": 39,
#             "cntrl_name": "LSB_for_controller_7"
#         },
#         "40": {
#             "cntrl_id": 40,
#             "cntrl_name": "LSB_for_controller_8"
#         },
#         "41": {
#             "cntrl_id": 41,
#             "cntrl_name": "LSB_for_controller_9"
#         },
#         "42": {
#             "cntrl_id": 42,
#             "cntrl_name": "LSB_for_controller_10"
#         },
#         "43": {
#             "cntrl_id": 43,
#             "cntrl_name": "LSB_for_controller_11"
#         },
#         "44": {
#             "cntrl_id": 44,
#             "cntrl_name": "LSB_for_controller_12"
#         },
#         "45": {
#             "cntrl_id": 45,
#             "cntrl_name": "LSB_for_controller_13"
#         },
#         "46": {
#             "cntrl_id": 46,
#             "cntrl_name": "LSB_for_controller_14"
#         },
#         "47": {
#             "cntrl_id": 47,
#             "cntrl_name": "LSB_for_controller_15"
#         },
#         "48": {
#             "cntrl_id": 48,
#             "cntrl_name": "LSB_for_controller_16"
#         },
#         "49": {
#             "cntrl_id": 49,
#             "cntrl_name": "LSB_for_controller_17"
#         },
#         "50": {
#             "cntrl_id": 50,
#             "cntrl_name": "LSB_for_controller_18"
#         },
#         "51": {
#             "cntrl_id": 51,
#             "cntrl_name": "LSB_for_controller_19"
#         },
#         "52": {
#             "cntrl_id": 52,
#             "cntrl_name": "LSB_for_controller_20"
#         },
#         "53": {
#             "cntrl_id": 53,
#             "cntrl_name": "LSB_for_controller_21"
#         },
#         "54": {
#             "cntrl_id": 54,
#             "cntrl_name": "LSB_for_controller_22"
#         },
#         "55": {
#             "cntrl_id": 55,
#             "cntrl_name": "LSB_for_controller_23"
#         },
#         "56": {
#             "cntrl_id": 56,
#             "cntrl_name": "LSB_for_controller_24"
#         },
#         "57": {
#             "cntrl_id": 57,
#             "cntrl_name": "LSB_for_controller_25"
#         },
#         "58": {
#             "cntrl_id": 58,
#             "cntrl_name": "LSB_for_controller_26"
#         },
#         "59": {
#             "cntrl_id": 59,
#             "cntrl_name": "LSB_for_controller_27"
#         },
#         "60": {
#             "cntrl_id": 60,
#             "cntrl_name": "LSB_for_controller_28"
#         },
#         "61": {
#             "cntrl_id": 61,
#             "cntrl_name": "LSB_for_controller_29"
#         },
#         "62": {
#             "cntrl_id": 62,
#             "cntrl_name": "LSB_for_controller_30"
#         },
#         "63": {
#             "cntrl_id": 63,
#             "cntrl_name": "LSB_for_controller_31"
#         },
#         "64": {
#             "cntrl_id": 64,
#             "cntrl_name": "damper_pedal"
#         },
#         "65": {
#             "cntrl_id": 65,
#             "cntrl_name": "portamento"
#         },
#         "66": {
#             "cntrl_id": 66,
#             "cntrl_name": "sostenuto"
#         },
#         "67": {
#             "cntrl_id": 67,
#             "cntrl_name": "soft_pedal"
#         },
#         "68": {
#             "cntrl_id": 68,
#             "cntrl_name": "legato_footswitch"
#         },
#         "69": {
#             "cntrl_id": 69,
#             "cntrl_name": "hold"
#         },
#         "70": {
#             "cntrl_id": 70,
#             "cntrl_name": "sound_controller_0"
#         },
#         "71": {
#             "cntrl_id": 71,
#             "cntrl_name": "sound_controller_1"
#         },
#         "72": {
#             "cntrl_id": 72,
#             "cntrl_name": "sound_controller_2"
#         },
#         "73": {
#             "cntrl_id": 73,
#             "cntrl_name": "sound_controller_3"
#         },
#         "74": {
#             "cntrl_id": 74,
#             "cntrl_name": "sound_controller_4"
#         },
#         "75": {
#             "cntrl_id": 75,
#             "cntrl_name": "sound_controller_5"
#         },
#         "76": {
#             "cntrl_id": 76,
#             "cntrl_name": "sound_controller_6"
#         },
#         "77": {
#             "cntrl_id": 77,
#             "cntrl_name": "sound_controller_7"
#         },
#         "78": {
#             "cntrl_id": 78,
#             "cntrl_name": "sound_controller_8"
#         },
#         "79": {
#             "cntrl_id": 79,
#             "cntrl_name": "sound_controller_9"
#         },
#         "80": {
#             "cntrl_id": 80,
#             "cntrl_name": "post_general_purpose_controller_0"
#         },
#         "81": {
#             "cntrl_id": 81,
#             "cntrl_name": "post_general_purpose_controller_1"
#         },
#         "82": {
#             "cntrl_id": 82,
#             "cntrl_name": "post_general_purpose_controller_2"
#         },
#         "83": {
#             "cntrl_id": 83,
#             "cntrl_name": "post_general_purpose_controller_3"
#         },
#         "84": {
#             "cntrl_id": 84,
#             "cntrl_name": "portamento_control"
#         },
#         "91": {
#             "cntrl_id": 91,
#             "cntrl_name": "effects_depth_0"
#         },
#         "92": {
#             "cntrl_id": 92,
#             "cntrl_name": "effects_depth_1"
#         },
#         "93": {
#             "cntrl_id": 93,
#             "cntrl_name": "effects_depth_2"
#         },
#         "94": {
#             "cntrl_id": 94,
#             "cntrl_name": "effects_depth_3"
#         },
#         "95": {
#             "cntrl_id": 95,
#             "cntrl_name": "effects_depth_4"
#         },
#         "96": {
#             "cntrl_id": 96,
#             "cntrl_name": "data_increment"
#         },
#         "97": {
#             "cntrl_id": 97,
#             "cntrl_name": "data_decrement"
#         },
#         "98": {
#             "cntrl_id": 98,
#             "cntrl_name": "nonreg_param_num_lsb"
#         },
#         "99": {
#             "cntrl_id": 99,
#             "cntrl_name": "nonreg_param_num_msb"
#         },
#         "100": {
#             "cntrl_id": 100,
#             "cntrl_name": "reg_param_num_lsb"
#         },
#         "101": {
#             "cntrl_id": 101,
#             "cntrl_name": "reg_param_num_msb"
#         },
#         "121": {
#             "cntrl_id": 121,
#             "cntrl_name": "mode_messages_0"
#         },
#         "122": {
#             "cntrl_id": 122,
#             "cntrl_name": "mode_messages_1"
#         },
#         "123": {
#             "cntrl_id": 123,
#             "cntrl_name": "mode_messages_2"
#         },
#         "124": {
#             "cntrl_id": 124,
#             "cntrl_name": "mode_messages_3"
#         },
#         "125": {
#             "cntrl_id": 125,
#             "cntrl_name": "mode_messages_4"
#         },
#         "126": {
#             "cntrl_id": 126,
#             "cntrl_name": "mode_messages_5"
#         },
#         "127": {
#             "cntrl_id": 127,
#             "cntrl_name": "mode_messages_6"
#         }
#     },
#     "12": {
#         "id": 12,
#         "name": "program_change",
#         "length": 1,
#         "params": [
#             "program_number"
#         ],
#         "mask": 127
#     },
#     "13": {
#         "id": 13,
#         "name": "channel_after_touch",
#         "length": 1,
#         "params": [
#             "amount"
#         ],
#         "mask": 127
#     },
#     "14": {
#         "id": 14,
#         "name": "pitch_bend",
#         "length": 2,
#         "params": [
#             "vlsb",
#             "vmsb"
#         ],
#         "mask": 127
#     }
# }

X = {
    8: {'id': 8,
        'length': 2,
        'mask': 127,
        'name': 'note_off',
        'params': ('note_number', 'velocity')},
    9: {'id': 9,
        'length': 2,
        'mask': 127,
        'name': 'note_on',
        'params': ('note_number', 'velocity')},
    10: {'id': 10,
        'length': 2,
        'mask': 127,
        'name': 'after_touch',
        'params': ('note_number', 'amount')},
    11: {0: {'cntrl_id': 0, 'cntrl_name': 'blank_select'},
        1: {'cntrl_id': 1, 'cntrl_name': 'modulation'},
        2: {'cntrl_id': 2, 'cntrl_name': 'breath_controller'},
        4: {'cntrl_id': 4, 'cntrl_name': 'foot_controller'},
        5: {'cntrl_id': 5, 'cntrl_name': 'portamento_time'},
        6: {'cntrl_id': 6, 'cntrl_name': 'data_entry'},
        7: {'cntrl_id': 7, 'cntrl_name': 'main_volume'},
        8: {'cntrl_id': 8, 'cntrl_name': 'balance'},
        10: {'cntrl_id': 10, 'cntrl_name': 'pan'},
        11: {'cntrl_id': 11, 'cntrl_name': 'expression_controller'},
        12: {'cntrl_id': 12, 'cntrl_name': 'effect_control_0'},
        13: {'cntrl_id': 13, 'cntrl_name': 'effect_control_1'},
        16: {'cntrl_id': 16, 'cntrl_name': 'general_purpose_controller_0'},
        17: {'cntrl_id': 17, 'cntrl_name': 'general_purpose_controller_1'},
        18: {'cntrl_id': 18, 'cntrl_name': 'general_purpose_controller_2'},
        19: {'cntrl_id': 19, 'cntrl_name': 'general_purpose_controller_3'},
        32: {'cntrl_id': 32, 'cntrl_name': 'LSB_for_controller_0'},
        33: {'cntrl_id': 33, 'cntrl_name': 'LSB_for_controller_1'},
        34: {'cntrl_id': 34, 'cntrl_name': 'LSB_for_controller_2'},
        35: {'cntrl_id': 35, 'cntrl_name': 'LSB_for_controller_3'},
        36: {'cntrl_id': 36, 'cntrl_name': 'LSB_for_controller_4'},
        37: {'cntrl_id': 37, 'cntrl_name': 'LSB_for_controller_5'},
        38: {'cntrl_id': 38, 'cntrl_name': 'LSB_for_controller_6'},
        39: {'cntrl_id': 39, 'cntrl_name': 'LSB_for_controller_7'},
        40: {'cntrl_id': 40, 'cntrl_name': 'LSB_for_controller_8'},
        41: {'cntrl_id': 41, 'cntrl_name': 'LSB_for_controller_9'},
        42: {'cntrl_id': 42, 'cntrl_name': 'LSB_for_controller_10'},
        43: {'cntrl_id': 43, 'cntrl_name': 'LSB_for_controller_11'},
        44: {'cntrl_id': 44, 'cntrl_name': 'LSB_for_controller_12'},
        45: {'cntrl_id': 45, 'cntrl_name': 'LSB_for_controller_13'},
        46: {'cntrl_id': 46, 'cntrl_name': 'LSB_for_controller_14'},
        47: {'cntrl_id': 47, 'cntrl_name': 'LSB_for_controller_15'},
        48: {'cntrl_id': 48, 'cntrl_name': 'LSB_for_controller_16'},
        49: {'cntrl_id': 49, 'cntrl_name': 'LSB_for_controller_17'},
        50: {'cntrl_id': 50, 'cntrl_name': 'LSB_for_controller_18'},
        51: {'cntrl_id': 51, 'cntrl_name': 'LSB_for_controller_19'},
        52: {'cntrl_id': 52, 'cntrl_name': 'LSB_for_controller_20'},
        53: {'cntrl_id': 53, 'cntrl_name': 'LSB_for_controller_21'},
        54: {'cntrl_id': 54, 'cntrl_name': 'LSB_for_controller_22'},
        55: {'cntrl_id': 55, 'cntrl_name': 'LSB_for_controller_23'},
        56: {'cntrl_id': 56, 'cntrl_name': 'LSB_for_controller_24'},
        57: {'cntrl_id': 57, 'cntrl_name': 'LSB_for_controller_25'},
        58: {'cntrl_id': 58, 'cntrl_name': 'LSB_for_controller_26'},
        59: {'cntrl_id': 59, 'cntrl_name': 'LSB_for_controller_27'},
        60: {'cntrl_id': 60, 'cntrl_name': 'LSB_for_controller_28'},
        61: {'cntrl_id': 61, 'cntrl_name': 'LSB_for_controller_29'},
        62: {'cntrl_id': 62, 'cntrl_name': 'LSB_for_controller_30'},
        63: {'cntrl_id': 63, 'cntrl_name': 'LSB_for_controller_31'},
        64: {'cntrl_id': 64, 'cntrl_name': 'damper_pedal'},
        65: {'cntrl_id': 65, 'cntrl_name': 'portamento'},
        66: {'cntrl_id': 66, 'cntrl_name': 'sostenuto'},
        67: {'cntrl_id': 67, 'cntrl_name': 'soft_pedal'},
        68: {'cntrl_id': 68, 'cntrl_name': 'legato_footswitch'},
        69: {'cntrl_id': 69, 'cntrl_name': 'hold'},
        70: {'cntrl_id': 70, 'cntrl_name': 'sound_controller_0'},
        71: {'cntrl_id': 71, 'cntrl_name': 'sound_controller_1'},
        72: {'cntrl_id': 72, 'cntrl_name': 'sound_controller_2'},
        73: {'cntrl_id': 73, 'cntrl_name': 'sound_controller_3'},
        74: {'cntrl_id': 74, 'cntrl_name': 'sound_controller_4'},
        75: {'cntrl_id': 75, 'cntrl_name': 'sound_controller_5'},
        76: {'cntrl_id': 76, 'cntrl_name': 'sound_controller_6'},
        77: {'cntrl_id': 77, 'cntrl_name': 'sound_controller_7'},
        78: {'cntrl_id': 78, 'cntrl_name': 'sound_controller_8'},
        79: {'cntrl_id': 79, 'cntrl_name': 'sound_controller_9'},
        80: {'cntrl_id': 80, 'cntrl_name': 'post_general_purpose_controller_0'},
        81: {'cntrl_id': 81, 'cntrl_name': 'post_general_purpose_controller_1'},
        82: {'cntrl_id': 82, 'cntrl_name': 'post_general_purpose_controller_2'},
        83: {'cntrl_id': 83, 'cntrl_name': 'post_general_purpose_controller_3'},
        84: {'cntrl_id': 84, 'cntrl_name': 'portamento_control'},
        91: {'cntrl_id': 91, 'cntrl_name': 'effects_depth_0'},
        92: {'cntrl_id': 92, 'cntrl_name': 'effects_depth_1'},
        93: {'cntrl_id': 93, 'cntrl_name': 'effects_depth_2'},
        94: {'cntrl_id': 94, 'cntrl_name': 'effects_depth_3'},
        95: {'cntrl_id': 95, 'cntrl_name': 'effects_depth_4'},
        96: {'cntrl_id': 96, 'cntrl_name': 'data_increment'},
        97: {'cntrl_id': 97, 'cntrl_name': 'data_decrement'},
        98: {'cntrl_id': 98, 'cntrl_name': 'nonreg_param_num_lsb'},
        99: {'cntrl_id': 99, 'cntrl_name': 'nonreg_param_num_msb'},
        100: {'cntrl_id': 100, 'cntrl_name': 'reg_param_num_lsb'},
        101: {'cntrl_id': 101, 'cntrl_name': 'reg_param_num_msb'},
        121: {'cntrl_id': 121, 'cntrl_name': 'mode_messages_0'},
        122: {'cntrl_id': 122, 'cntrl_name': 'mode_messages_1'},
        123: {'cntrl_id': 123, 'cntrl_name': 'mode_messages_2'},
        124: {'cntrl_id': 124, 'cntrl_name': 'mode_messages_3'},
        125: {'cntrl_id': 125, 'cntrl_name': 'mode_messages_4'},
        126: {'cntrl_id': 126, 'cntrl_name': 'mode_messages_5'},
        127: {'cntrl_id': 127, 'cntrl_name': 'mode_messages_6'},
        'id': 11,
        'length': 2,
        'mask': 127,
        'params': ['controller_type', 'value']},
    12: {'id': 12,
        'length': 1,
        'mask': 127,
        'name': 'program_change',
        'params': ('program_number',)},
    13: {'id': 13,
        'length': 1,
        'mask': 127,
        'name': 'channel_after_touch',
        'params': ('amount',)},
    14: {'id': 14,
        'length': 2,
        'mask': 127,
        'name': 'pitch_bend',
        'params': ('vlsb', 'vmsb')
        }
    }


from rmidi.constant import Constants
import os
import json
def dict_to_event_json(filename, dictn):
    with open(filename, 'w') as wrt:
        str_ = json.dumps(dictn, indent=4)
        wrt.write(str_)

def to_channel_event_dict():
    '''
        id : Midi Channel Event Id
        name : Corresponding Midi Event Name
        length : length of parameters needed
        params : list of parameter values, Big-Endian format; 0 indicates most significant bit
        mask : bitmask over the 8 bit value; It can be list or sin 
    '''
    ch_dict = {}
    key = ['id', 'name', 'length', 'params', 'mask']
    for row in Constants.ch_event_format:
        rw0 = row[0]
        if rw0 == 0xB:
            ch_dict[rw0] = to_ch_controller_dict()
            continue
        ch_dict[rw0] = {}
        for i, k in enumerate(key[:3]):
            ch_dict[rw0][k] = row[i]
        ch_dict[rw0]['params'] = row[3: 3 + row[2]]
        ch_dict[rw0]['mask'] = 127 if row[3 + row[2]] == 'int' else row[3 + row[2]: ]

    return ch_dict

def to_meta_event_dict():
    '''
        id : Midi Channel Event Id, Always (0xFF)
        name : Corresponding Midi Event Name, Always meta
        type_id: Midi Meta Event Sub Type Id
        type_name: Midi Meta Event Sub Type Name
        length : length of parameters needed
        params : list of parameter values, Big-Endian format; 0 indicates most significant bit
        mask : bitmask over the 8 bit value; It can be list or sin 
    '''
    meta_dict = {0xFF : {'id' : 0xFF, 'name': 'meta'}}
    key = ['type_id', 'type_name', 'length', 'params', 'mask']
    for row in Constants.meta_event_format:
        rw0 = row[0]
        meta_dict[0xFF][rw0]= {}
        for i, k in enumerate(key[:3]):
            meta_dict[0xFF][rw0][k] = row[i]
        meta_dict[0xFF][rw0]['params'] = row[3: 3 + row[2]] if row[2] != 0 and row[2] != -1 else None if row[2] == 0 else 'text'
        meta_dict[0xFF][rw0]['dtype'] = row[4] if row[2] == 0 or row[2] == -1 else row[3 + row[2]] #  127 if row[3 + row[2]] == 'str' else None #row[3 + row[2]] if row[2] != 0 or row[2] != -1 else row[4]  
        meta_dict[0xFF][rw0]['mask'] = row[4 + row[2] : ] if row[3 + row[2]] == 'int' else None if row[4] == None else 127 # row[4 + row[2]: ]

    # print(meta_dict)
    return meta_dict

def to_sys_event_dict():
    sys_dict = {}
    key = ['id', 'name', 'length', 'params', 'dtype', 'mask']
    for e in Constants.sys_event_format:
        sys_dict[e[0]] = {}
        for i, k in enumerate(key):
            sys_dict[e[0]][k] = e[i] 
    # print(sys_dict)
    return sys_dict

def to_ch_controller_dict():
    '''
        cntrl_id : Midi Channel Controller Event Subtype Id
        cntrl_name : Corresponding Midi Controller Event Subtype Name
    '''
    cntrl = Constants.controller
    res = {
        "id": 0xB,
        "name": "controller",
        "length": 2,
        "params": [
            "controller_type",
            "value"
        ],
        "mask": 127
    }
    for cnt in cntrl:
        if cnt[2] == 1: 
            b = 1
            res[cnt[0]] = {'cntrl_id' : cnt[0], 'cntrl_name' : cnt[1]}  
        else :
            b = 0
            res[cnt[0]] = {}

        for ky in range(cnt[0] + b, cnt[0] + cnt[2]):
            res[ky] = {'cntrl_id' : ky, 'cntrl_name' : cnt[1] + '_' + str(ky - cnt[0])}

    # print(res)
    return res

def write_dict(dictn, filename, variable_name = 'X'):
    pass

# def ch_event_format():
#     f = './rmidi/constant/ch_event_format.json'
#     # f = './ch_event_format.json.py'
#     exists = os.path.exists(f)
#     with open('./rmidi/constant/ch_event_format.json', 'r') as f:
#         return json.load(f.read())

# def meta_event_format():
#     f = '.\\rmidi\\constant\\meta_event_format.json'
#     # f = './meta_event_format.json.py'
#     exists = os.path.exists(f)
#     print(exists, "   ", os.listdir())
#     with open(f, 'r+') as fil:
#     # st = fil.read()
#     # fil.close()
#         res = json.load(fil)


#     return res

# def sys_event_format():
#     # f = './rmidi/constant/'
#     with open('./rmidi/constant/sys_event_format.json.py', 'r') as f:
#         return json.load(f.read())

# ch_event = to_channel_event_dict()
# dict_to_event_json('./rmidi/constant/ch_event_format.json', ch_event)

# meta_event = to_meta_event_dict()
# dict_to_event_json('./rmidi/constant/meta_event_format.json', meta_event)

# # ch_cntrl_event = to_ch_controller_dict()
# # dict_to_event_json('./rmidi/constant/ch_controller.json', ch_cntrl_event)

# sys_event = to_sys_event_dict()
# dict_to_event_json('./rmidi/constant/sys_event_format.json', sys_event)